/*---------------------------------------------------------------------------*/
/* Hardware description of the EB285 board for Angel.
 *
 * $Revision: 1.1.1.1 $
 *   $Author: drusling $
 *     $Date: 1998/09/22 15:48:17 $
 *
 * Copyright Advanced RISC Machines Limited, 1995.
 *
 */
/*---------------------------------------------------------------------------*/

#ifndef __angel285_h
#define __angel285_h

/*---------------------------------------------------------------------------*/

#include        "platform.h"

#if (defined(__arm) && defined(DEBUG) && defined(PIDULATOR))
/* This uses a feature in the version of PIDulator being used to test
 * Angel. There is a SWI defined to enable strings to be output
 * directly from the emulator, without the SWI being passed through
 * the normal vector.
 */
void __swi(0xFFFE00) PIDprint(char *addr) ;
void __swi(0xFFFD00) PIDhex8(unsigned int val) ;
void __swi(0xFFFC00) PIDhex2(unsigned char val) ;
void __swi(0xFFFB00) PIDchar(char chr) ;
void __swi(0xFFFA00) PIDdecimal(unsigned int val) ;
#else /* no ARM Emulator debugging */
#ifdef DEBUG
#define PIDprint(a) PrintStringc2(a)
#define PIDhex8(a) PrintHexWordc2(a)
#define PIDhex2(a) PrintHexWordc2(a)
#define PIDchar(a)      /* do nothing */
#define PIDdecimal(a)   /* do nothing */
#else
/* We use single argument function prototypes because it is hard
 * constructing vararg macros :-)
 */
#define PIDprint(a)     /* do nothing */
#define PIDhex8(a)      /* do nothing */
#define PIDhex2(a)      /* do nothing */
#define PIDchar(a)      /* do nothing */
#define PIDdecimal(a)   /* do nothing */
#endif
#endif /* DEBUG */

/*---------------------------------------------------------------------------*/
/* This manifest is used by the generic Angel serial device driver
 * code. If it is defined then the code is constructed assuming a
 * synchronous (polled) serial system. If the manifest is *NOT*
 * defined, then the code assumes an asynchronous (interrupt driven)
 * serial world. This affects how the messages are routed to the
 * relevant channel handler.
 */
#if 0 /* 0 = interrupt driven world; 1 = polled world */
#define SERIAL_SYNC (1) /* We are starting with a simple polled world */
#endif

/*---------------------------------------------------------------------------*/
/*-- system support macros --------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/* These macros are the hi-level 'C' equivalents of the respective
 * assembler versions.
 */

/* SYSTRACE can be used to provide some form of LSA triggering, or
 * debug messages via some private I/O channel. The "id" and "value"
 * parameters are private parameters to the macro. i.e. the "id"
 * parameter need not result in a specific trace value, but can be a
 * modifier to the current trace state. e.g. when dealing with LEDs it
 * can be a command to change a particular bit, rather than the
 * complete bank of LEDs.
 */
#define SYSTRACE(id,value) /* do nothing */

/*---------------------------------------------------------------------------*/
/* The following manifest is used to construct the (build-time fixed
 * size) Angel interrupt vector. It should be set to the number of
 * interrupt sources that will be handled by the target Angel
 * version. For the VY86PID implementation defined by this header, we
 * only support the serial interrupt. All other sources are left for
 * the user application or O/S loaded.
 */
#define NUMINTS   (1)   /* Number of Angel interrupt vectors */
#define SERIALIDX (0)   /* Index in the NUMINTS vector for the serial handler */

/*---------------------------------------------------------------------------*/

/*
 * Definitions for the UART (used by serial.c).
 */

/*
 * Register offsets.
 */
#define UART_RX             0      /* Receive port, read only */
#define UART_TX             0      /* transmit port, write only */
#define UART_IER            1      /* Interrupt enable, read/write */
#define UART_IIR            2      /* Interrupt id, read only */
#define UART_FIFO_CONTROL   2      /* FIFO control, write only */
#define UART_LCR            3      /* Line control register */
#define UART_MCR            4      /* Modem control register */
#define UART_LSR            5      /* Line Status register */
#define UART_MSR            6      /* Modem Status register */

/* with the DLAB bit set, the first two registers contain the baud rate */
#define UART_DLLSB          0
#define UART_DLMSB          1

/*
 * Line control register 
 */
#define LCR_DB              3      /* Data bits in transmission (0 = 5, 1 = 6, 2 = 7, 3 = 8) */
#define LCR_SB              4      /* Stop bit */
#define LCR_PE              8      /* Parity enable */
#define LCR_EP             16      /* Even parity */
#define LCR_SP             32      /* Stick parity */
#define LCR_BC             64      /* break control */
#define LCR_DLAB          128      /* Divisor latch access bit */


/*
 *  Modem Control register
 */
#define MCR_IRQ_ENABLE       8

/*
 * Line status bits.
 */
#define LSR_DR               1     /* receive data ready */
#define LSR_OE               2     /* overrun error */
#define LSR_PE               4     /* parity error */
#define LSR_FE               8     /* Framing error */
#define LSR_BI              16     /* Break interrupt */
#define LSR_THRE            32     /* Ready to transmit */
#define LSR_TEMT            64     /* Transmitter empty */

#define LSR_ERROR           (LSR_OE | LSR_PE | LSR_FE) 

/*
 * Interrupt Identification register (IIR)
 */
#define IIR_IP               1     /* No Interrupt pending */
#define IIR_HPIP             6     /* Highest priority interrupt pending */

/*
 * Interrupt enable register (IER)
 */
#define IER_RDA              1     /* Received data available */
#define IER_THRE             2     /* Transmitter holding register empty */
#define IER_RLS              4     /* Receiver line status */
#define IER_MS               8     /* Modem status */

/*---------------------------------------------------------------------------
 * Definitions for XBUS UART. Only define as standard definitions if
 * UARTS != 0. Must be globally accessable for debug when using 21285 UART.
 */

#define LOG_READ_INTERRUPT(p)   inb( (void *)(p + UART_IIR) )
/* Enable & disable both write the value to the same address */
#define LOG_ENABLE_INT(v, b, p) outb(v, (void *)((p) + UART_IER))
#define LOG_DISABLE_INT(v, b, p) outb(v, (void *)((p) + UART_IER))
#define LOG_ENABLE_INTERRUPT(p) outb(MCR_IRQ_ENABLE, (void *)((p) + UART_MCR) )

#define LOG_NO_INTERRUPT        IIR_IP
#define LOG_TX_INTERRUPT        0x02
#define LOG_RX_INTERRUPT        0x04

#define LOG_TX_ENABLE           IER_THRE
#define LOG_RX_ENABLE           (IER_RDA + IER_RLS)

#define LOG_GET_STATUS(p)       (xinb( (void *)((p) + UART_LSR) ) & 0xFF)
#define LOG_GET_CHAR(p)         (xinb((void *)((p) + UART_RX)))
#define LOG_PUT_CHAR(p, c)      (xoutb(c, (p) + UART_TX))
#define LOG_TX_EMPTY(a)         (xinb( (void *)((a) + UART_LSR) ) & LSR_TEMT)
#define LOG_TX_READY(a)         ((a) & LSR_THRE)
#define LOG_RX_DATA(a)          ((a) & LSR_DR)

#if UARTS != 0
#define READ_INTERRUPT(p)       LOG_READ_INTERRUPT(p)
/* Enable & disable both write the value to the same address */
#define ENABLE_INT(v, b, p)     LOG_ENABLE_INT(v, b, p) 
#define DISABLE_INT(v, b, p)    LOG_DISABLE_INT(v, b, p)        
#define ENABLE_INTERRUPT(p)     LOG_ENABLE_INTERRUPT(p) 

#define NO_INTERRUPT            LOG_NO_INTERRUPT                
#define TX_INTERRUPT            LOG_TX_INTERRUPT                
#define RX_INTERRUPT            LOG_RX_INTERRUPT                

#define TX_ENABLE               LOG_TX_ENABLE           
#define RX_ENABLE               LOG_RX_ENABLE           

#define GET_STATUS(p)           LOG_GET_STATUS(p)               
#define GET_CHAR(p)             LOG_GET_CHAR(p)         
#define PUT_CHAR(p, c)          LOG_PUT_CHAR(p, c)              
#define TX_EMPTY(a)             LOG_TX_EMPTY(a)         
#define TX_READY(a)             LOG_TX_READY(a)         
#define RX_DATA(a)              LOG_RX_DATA(a)          
#endif

extern int inb(void *p) ;
extern int outb(unsigned char b, void *p) ;

#endif /* __angel285_h */


