/*
 * Copyright (C) 1996
 * Digital Equipment Corporation, Maynard, Massachusetts.
 * All Rights Reserved
 *
 * Permission to  use, copy,  modify, and distribute  this software
 * and its documentation for any  purpose and without fee is hereby 
 * granted, provided that the  copyright notice and this permission 
 * notice   appear  in  all   copies  of  software  and  supporting  
 * documentation,  and  that  the  name  of Digital  not be used in 
 * advertising or  publicity  pertaining to   distribution  of  the 
 * software  without  specific,  written prior  permission. Digital  
 * grants  this  permission  provided  that you  prominently  mark, 
 * as  not part  of the original,  any modifications that  is  made  
 * to this software or documentation.
 *
 * Digital Equipment Corporation  disclaims all  warranties  and/or 
 * guarantees with  regard to  this software, including all implied 
 * warranties   of  fitness  for   a   particular    purpose    and   
 * merchantibility, and makes no representations regarding  use of,  
 * or  the  results of  the use of, the  software and documentation 
 * in terms  of correctness, accuracy, reliability, currentness, or 
 * otherwise;  and   you rely  on  the software,  documentation and 
 * results solely at your own risk.
 *
 * ======================================================================
 *
 * Abstract:	Device node tree routines
 *
 *	$Id: devnode.c,v 1.1 1999/01/26 14:16:15 mquinn Exp $
 *
 * ====================================================================== */
#include "system.h"
#include "pci.h"

static int devNodesInitialized = FALSE ;
static devNodeType StaticDevNodes[MAX_DEVNODE];
static devNodeType *freeNodes = NULL ;

devNodeType *devNodes = NULL ;

/*============================================================================
 *
 *  routine:     SAir_AllocateDevNode
 *
 *  parameters:  void
 *
 *  description: Allocate new devnode structure.
 *
 *  calls:       None.
 *
 *  returns:     Pointer to a devNodeType or NULL if it fails.
 *
 */

devNodeType *SAir_AllocateDevNode(void)
{
devNodeType	*newNode ;

	newNode = freeNodes ;
	if (freeNodes != NULL) {
		freeNodes = freeNodes->next ;
		newNode->next = NULL ;
	}

	return newNode ;
}

/*============================================================================
 *
 *  routine:     addToQueue(devNodeType **queue, devNodeType newNode)
 *
 *  parameters:  void
 *
 *  description: Insert a filled out devNode into a devNode list.
 *
 *  calls:       None.
 *
 *  returns:     None.
 *
 */
#define __addToQueue(q,n) {(n)->next = (q); (q) = (n);}

/*============================================================================
 *
 *  routine:     InitDevNodes
 *
 *  parameters:  void
 *
 *  description: Initializes the devNode data structures.
 *
 *  calls:       None.
 *
 *  returns:     None.
 *
 */
static void initDevNodes(void)
{
int		 i ;
devNodeType	*newNode ;

	for (i = 0; i < MAX_DEVNODE; i++) {
		newNode = &StaticDevNodes[i] ;
		__addToQueue(freeNodes, newNode) ;
	}
	devNodesInitialized = TRUE ;
}

/*============================================================================
 *
 *  routine:     SAr_BuildDevNodeTree(void)
 *
 *  parameters:  void
 *
 *  description: this routine will insert all known fixed devices into the
 *               device node tree. It will then scan the PCI buses looking
 *               for valid devices.
 *
 *  calls:       addDevNode(), pciCfgReadLong()
 *
 *  returns:     status of the devNode tree build
 *
 */

S8 SAr_BuildDevNodeTree(void)
{
#if uHAL_PCI != 0
U8	bus, slot, func = 0 ;
U32	vendor, handle ;
S32	status ;
#endif

	/* if we haven't initialized, then do so */
	if (!devNodesInitialized) initDevNodes() ;

	/* add all the fixed devices on the platform into the devNode tree */
	SAir_AddFixedDevNodes() ;

#if uHAL_PCI != 0
	if (SAr_PciHost()) {
		/* Our 21285 will be the first devnode - bus 0, slot 0 */
		/* now scan all possible buses and slots looking for devices */
		for (bus = 0; bus < MAX_PCI_BUS; bus++) {
		    for (slot = 0; slot < MAX_PCI_SLOT; slot++) {
			/* make handle for device at this slot */
			handle = MAKE_HANDLE(PCI_TYPE, bus, slot, func) ;

			/* read the device configuration space */
			vendor = SAr_GetPciVendorID(handle, &status);

			/* check if it responds */
			if (vendor != PCI_NOTFITTED) {
				/* pci drivers are required to read the
				 * BARs for base addr
				 */ 
				vendor <<= 16;
				vendor |= SAr_GetPciDeviceID(handle, &status);
				SAr_AddDevNode(PCI_TYPE, vendor, handle, 0) ;
			}
		    }
		}
	}
#endif
	return MSG_SUCCESS;
}

/*============================================================================
 *
 *  routine:     SAr_PrintDevNodeTree(void)
 *
 *  parameters:  void
 *
 *  description: this routine displays the current devNode tree.
 *
 *  calls:       
 *
 *  returns:     None.
 *
 */

void SAr_PrintDevNodeTree(void)
{
devNodeType	*current = devNodes ;

	if (current == NULL)
		SAr_printf("No devNodes defined\n") ;

	while (current != NULL) {
		SAr_printf("devNode:\n") ;
		SAr_printf("  id:     0x%08X\n", current->id) ;
		SAr_printf("  handle: 0x%08X\n", current->handle) ;
		SAr_printf("  intDma: 0x%08X\n", current->intDma) ;
		current = current->next ;
	}
}


/*============================================================================
 *
 *  routine:     SAr_AddDevNode(U32 type, U32 id, U32 handle, U32 intDma)
 *
 *  parameters:  type   = bus type (0=PCI,1=xbus,2=CMA bus)
 *               id     = id used to uniquely identify device in devNode tree
 *               handle = handle of device to add
 *               intDma = interrupt information and DMA channel assigned to 
 *                        device
 *                                                    
 *  description: this routine will insert this device into the devNode tree. 
 *               It will keep track of available devNodes. 
 *
 *  calls:       none
 *
 *  returns:     void
 *
 */
void SAr_AddDevNode(U32 type, U32 id, U32 handle, U32 intDma)
{
devNodeType	*newNode ;

	/* first allocate a new one */
	newNode = SAir_AllocateDevNode() ;
	if (newNode != NULL) {
		newNode->id = id ;
		newNode->handle = handle | type ;
		newNode->intDma = intDma ;
		__addToQueue(devNodes, newNode) ;
	}
}       


/*============================================================================
 *
 *  routine:     SAr_GetDevNodeHandle(U32 id, U32 index)
 *
 *  parameters:  id     = id used to uniquely identify device in devNode tree
 *               index  = iteration of device in the devNode tree. This is used
 *                        if there is more than one instance of a device. 
 *                                                    
 *  description: this routine will search for a valid device with a matching
 *               id in the devNode tree.
 *
 *  calls:       none
 *
 *  returns:     handle for the device if found, returns 0 if failed.
 *
 */

U32 SAr_GetDevNodeHandle(U32 id, U32 index)
{
devNodeType	*node ;
U32		 numDev = 0 ;

	for (node = devNodes ; node != NULL ; node = node->next) {
		if (node->id == id) {
			if (numDev++ == index) {
				return node->handle ;
			}
		}
	}

	/*  failure to find device in devNode tree */
	return 0 ;
}       

/*============================================================================
 *
 * routine:	SAr_FindDevNode(U32 handle)
 *
 * parameters:	handle = handle used to uniquely identify device in tree
 *
 * description:	this routine will search for a valid device with a matching
 *		handle in the devNode tree.
 *
 * calls:	none
 *
 * returns:	handle for the device if found, returns 0 if failed.
 *
 */

devNodeType *SAr_FindDevNode(U32 handle)
{
devNodeType	*node ;

	/* Don't waste time on bad handles */
	if (handle == 0)
		return NULL ;

	for (node = devNodes ; node != NULL ; node = node->next) {
		if (node->handle == handle) {
			return node ;
		}
	}

	/*  failure to find device in devNode tree */
	return NULL ;
}       

