/*
 * Copyright (C) 1996
 * Digital Equipment Corporation, Maynard, Massachusetts.
 * All Rights Reserved
 *
 * Permission to  use, copy,  modify, and distribute  this software
 * and its documentation for any  purpose and without fee is hereby 
 * granted, provided that the  copyright notice and this permission 
 * notice   appear  in  all   copies  of  software  and  supporting  
 * documentation,  and  that  the  name  of Digital  not be used in 
 * advertising or  publicity  pertaining to   distribution  of  the 
 * software  without  specific,  written prior  permission. Digital  
 * grants  this  permission  provided  that you  prominently  mark, 
 * as  not part  of the original,  any modifications that  is  made  
 * to this software or documentation.
 *
 * Digital Equipment Corporation  disclaims all  warranties  and/or 
 * guarantees with  regard to  this software, including all implied 
 * warranties   of  fitness  for   a   particular    purpose    and   
 * merchantibility, and makes no representations regarding  use of,  
 * or  the  results of  the use of, the  software and documentation 
 * in terms  of correctness, accuracy, reliability, currentness, or 
 * otherwise;  and   you rely  on  the software,  documentation and 
 * results solely at your own risk.
 *
 * ======================================================================
 *
 * Abstract:	PCI configuration space access routines
 *
 *	$Id: pci.c,v 1.1 1999/01/26 14:16:34 mquinn Exp $
 *
 * ====================================================================== */
 
#include "system.h"


#if uHAL_PCI != 0

#include "pci.h"

/*============================================================================
 *
 * routine:	SAir_CheckCfgParams(U32 devHandle, U32 offset)
 *
 * parameters:	handle = handle of device that we are assigning resources to
 *		offset = offset into PCI config space for current BAR
 *
 * description:	This routine checks that all the parameters fit into the
 *		PCI config space.
 *
 * calls:	none
 *
 * returns:	SUCCESS or ERROR_STATUS
 */

static U32 SAir_CheckCfgParams(U32 devHandle, U32 offset)
{
U32	bus, device, func ;

	/* extract the bus, device, and function number from the handle */
	bus = (U32)GET_BUS(devHandle) ;
	device = (U32)GET_DEVICE(devHandle) ;
	func = (U32)GET_FUNCTION(devHandle) ;

	/* check that all parameters are valid */
	if (bus >= 256)
		return(MSG_ERROR_BAD_BUS) ; 
	if (device >= 12)
		return(MSG_ERROR_BAD_DEVICE) ; 
	if (func >= 8)
		return(MSG_ERROR_BAD_FUNC) ; 
	if (offset >= 256)
		return(MSG_ERROR_BAD_OFFSET) ; 

	return(MSG_SUCCESS) ;
}

/*============================================================================
 *
 * routine:	SAr_PciCfgReadByte(U32 handle, U32 offset, PS32 pStatus)
 * 		SAr_PciCfgReadShort(U32 handle, U32 offset, PS32 pStatus)
 * 		SAr_PciCfgReadLong(U32 handle, U32 offset, PS32 pStatus)
 *
 * parameters:	handle = handle of device that we are assigning resources to
 *		offset = offset into PCI config space for current BAR
 *		data    = value to be written
 *		pStatus = SUCCESS/FAIL flag
 *
 * description:	These routines read a byte, short or long value from the given
 *		PCI config register.
 *
 * calls:	SAir_CheckCfgParams(), SAir_PciMakeConfigAddress()
 *
 * returns:	data - value read.
 */

U8 SAr_PciCfgReadByte(U32 devHandle, U32 offset, PS32 pStatus)
{
PU8	pAddress ;
U8	data ;

	/* check device params */
	*pStatus = (S32)SAir_CheckCfgParams(devHandle, offset) ;
	if (*pStatus != MSG_SUCCESS)
		return(0) ;

	/* generate the address of correct configuration space */
	pAddress = (PU8)(SAir_PciMakeConfigAddress(devHandle, offset)) ; 

	/* now that we have valid params, go read the config space data */
	data = *pAddress ;
	*pStatus = MSG_SUCCESS ;

	return(data) ;
}

U16 SAr_PciCfgReadShort(U32 devHandle, U32 offset, PS32 pStatus)
{
PU16	pAddress ;
U16	data ;

	/* check device params */
	*pStatus = (S32)SAir_CheckCfgParams(devHandle, offset) ;
	if (*pStatus != MSG_SUCCESS)
		return(0) ;

	/* generate the address of correct configuration space */
	pAddress = (PU16)(SAir_PciMakeConfigAddress(devHandle, offset)) ; 

	/* now that we have valid params, go read the config space data */
	data = *pAddress ;
	*pStatus = MSG_SUCCESS ;

	return(data) ;
}

U32 SAr_PciCfgReadLong(U32 devHandle, U32 offset, PS32 pStatus)
{
PU32	pAddress ;
U32	data ;

	/* check device params */
	*pStatus = (S32)SAir_CheckCfgParams(devHandle, offset) ;
	if (*pStatus != MSG_SUCCESS)
		return(0) ;

	/* generate the address of correct configuration space */
	pAddress = (PU32)(SAir_PciMakeConfigAddress(devHandle, offset)) ; 

	/* now that we have valid params, go read the config space data */
	data = *pAddress ;
	*pStatus = MSG_SUCCESS ;

	return(data) ;
}

/*============================================================================
 *
 * routine:	SAr_PciCfgReadByte(U32 handle, U32 offset, U32 data, PS32 pStatus)
 * 		SAr_PciCfgReadShort(U32 handle, U32 offset, U32 data, PS32 pStatus)
 * 		SAr_PciCfgReadLong(U32 handle, U32 offset, U32 data, PS32 pStatus)
 *
 * parameters:	handle  = handle of device that we are assigning resources to
 *		offset  = offset into PCI config space for current BAR
 *		data    = value to be written
 *		pStatus = SUCCESS/FAIL flag
 *
 * description:	These routines write a byte, short or long value to the given
 *		PCI config register.
 *
 * calls:	SAir_CheckCfgParams(), SAir_PciMakeConfigAddress()
 *
 * returns:	none
 */

void SAr_PciCfgWriteByte(U32 devHandle, U32 offset, U32 data, PS32 pStatus)
{
PU8	pAddress ;

	/* check device params */
	*pStatus = (S32)SAir_CheckCfgParams(devHandle, offset) ;
	if (*pStatus != MSG_SUCCESS)
		return ;

	/* generate the address of correct configuration space */
	pAddress = (PU8)(SAir_PciMakeConfigAddress(devHandle, offset)) ; 

	/* now that we have valid params, go write the config space data */
	*pAddress = (U8)data ;
	*pStatus = MSG_SUCCESS ;

	return ;
}

void SAr_PciCfgWriteShort(U32 devHandle, U32 offset, U32 data, PS32 pStatus)
{
PU16	pAddress ;

	/* check device params */
	*pStatus = (S32)SAir_CheckCfgParams(devHandle, offset) ;
	if (*pStatus != MSG_SUCCESS)
		return ;

	/* generate the address of correct configuration space */
	pAddress = (PU16)(SAir_PciMakeConfigAddress(devHandle, offset)) ; 

	/* now that we have valid params, go write the config space data */
	*pAddress = (U16)data ;
	*pStatus = MSG_SUCCESS ;

	return ;
}

void SAr_PciCfgWriteLong(U32 devHandle, U32 offset, U32 data, PS32 pStatus)
{
PU32	pAddress ;

	/* check device params */
	*pStatus = (S32)SAir_CheckCfgParams(devHandle,offset) ;
	if (*pStatus != MSG_SUCCESS)
		return ;

	/* generate the address of correct configuration space */
	pAddress = (PU32)(SAir_PciMakeConfigAddress(devHandle, offset)) ; 

	/* now that we have valid params, go write the config space data */
	*pAddress = (U32)data ;
	*pStatus = MSG_SUCCESS ;

	return ;
}

/*============================================================================
 *
 * routine:	SAir_AlignAddress(U32 address, U32 boundary)
 *
 * parameters:	address = start address of the block to align
 *		boundary = alignment boundary (ie 4K = 4K alignment)
 *
 * description:	this routine will attempt to align the address up to 
 *		the boundary specified.
 *
 * calls:	none
 *
 * returns:	aligned address
 */

static U32 SAir_AlignAddress (U32 address, U32 boundary)
{
	if ((address % boundary) == 0)
		return address ;

	return (U32)(address + (boundary - (address % boundary))) ;
}

static U32	SAiv_PciMemAvailableAddress = SZ_1M ;
static U32	SAiv_PciIoAvailableAddress = 0x400 ;

/*============================================================================
 *
 * routine:	SAir_GetNextPciAddr(U32 *base, U32 size)
 *
 * parameters:	*base - pointer to next available PCI address
 *		size  - amount of memory to be assigned
 *
 * description:	this routine will return the next available aligned memory 
 *		address block. This address block is added to PCI_MEM to
 *		obtain the memory location as seen by the StrongARM.
 *
 * calls:	SAir_AlignAddress()
 *
 * returns:	available aligned memory address
 */

U32 SAir_GetNextPciAddr(U32 *base, U32 size)
{
U32	address ;

	/* align the address to size (ie 4 meg size requires 4 meg alignment */
	address = SAir_AlignAddress(*base, size) ;

	/* generate next available address for next call */
	*base = address + size ;

#ifdef DEBUG
	SAr_printf("device requesting memory space %d bytes, assigned %08Xh\n",
		size, address) ;
#endif

	return(address) ;
}

/*============================================================================
 *
 * routine:	SAir_AssignMemAddr(U32 size)
 *
 * parameters:	size - converted value read from BAR after writing PCI_INVALID
 *
 * description:	this routine will return the next available aligned memory 
 *		address block. This address block is added to PCI_MEM to
 *		obtain the memory location as seen by the StrongARM.
 *
 * calls:	SAir_GetNextPciAddr()
 *
 * returns:	available aligned memory address
 */

U32 SAir_AssignMemAddr(U32 size)
{
	return(SAir_GetNextPciAddr(&SAiv_PciMemAvailableAddress, size)) ;
}

/*============================================================================
 *
 * routine:	SAir_AssignIoAddr(U32 size)
 *
 * parameters:	size - converted value read from BAR after writing PCI_INVALID
 *
 * description:	this routine will return the next available aligned IO 
 *		address block (must be < 1M, according to PCI spec.). This
 *		address is added to PCI_IO to obtain the memory location
 *		as seen by the StrongARM.
 *
 * calls:	SAir_GetNextPciAddr()
 *
 * returns:	available aligned IO address
 */

U32 SAir_AssignIoAddr(U32 size)
{
	return(SAir_GetNextPciAddr(&SAiv_PciIoAvailableAddress, size)) ;
}

/*============================================================================
 *
 * routine:	SAir_ReadBARSize(U32 handle, U32 offset)
 *
 * parameters:	handle = handle of device that we are assigning resources to
 *		offset = offset into PCI config space for current BAR
 *
 * description:	this routine will attempt to glean resource requirements
 *		from the device pointed to by handle
 *
 * calls:	pciCfgReadLong(), pciCfgWriteLong(),
 *
 * returns:	data - value read from BAR
 */

U32 SAir_ReadBARSize(U32 handle, U32 *sizePtr, U32 offset)
{
U32	data ;
S32	status ;

	/* write PCI_INVALID to BAR to get size */
	SAr_PciCfgWriteLong(handle, offset, PCI_INVALID, &status) ;
	/* check if this is an implemented BAR (size != 0) */
	data = SAr_PciCfgReadLong(handle, offset, &status) ;
	if (data != 0) {
		/* Keep a note of the size, because it isn't easy
		 * to read - have to disable card, read address,
		 * set to -1, read size, reload address & re-enable!
		 */
		*sizePtr = (~(data & 0xFFFFFFF0)) + 1 ;
	}
	else {
		*sizePtr = 0 ;
	}

	return (data) ;
}

/*============================================================================
 *
 * routine:	SAir_WriteBARSize(U32 handle, U32 data, U32 size, U32 offset)
 *
 * parameters:	handle = handle of device that we are assigning resources to
 *		data   = raw data read from BAR
 *		size   = size of PCI space required
 *		offset = offset into PCI config space for current BAR
 *
 * description:	this routine writes the base address for the requested size
 *		to the device pointed to by handle
 *
 * calls:	pciCfgWriteLong(),
 *		SAir_AssignIoAddr(), SAir_AssignMemAddr()
 *
 * returns:	void
 */

S32 SAir_WriteBARSize(U32 handle, U32 data, U32 size, U32 offset)
{
S32	status, barFlag = 0 ;

	/* Do not assign PCI space to ROM Expansion BAR to IO space */
	if ((offset != PCI_ROM_BAR) && (data & PCI_IO_ENABLE)) {
		/* IO space BAR */
		SAr_PciCfgWriteLong(handle, offset,
			SAir_AssignIoAddr(size), &status) ;
		barFlag |= PCI_IO_ENABLE ;
	}
	else {
		/* memory space BAR */
		SAr_PciCfgWriteLong(handle, offset,
			SAir_AssignMemAddr(size), &status) ;
		barFlag |= PCI_MEM_ENABLE ;
	}

	return barFlag ;
}

/*============================================================================
 *
 * routine:	SAir_AssignResources(U32 handle, U32 *sizePtr)
 *
 * parameters:	handle = handle of device that we are assigning resources to
 *		*sizePtr = pointer to an array which can hold the BAR sizes.
 *
 * description:	this routine will attempt to glean resource requirements
 *		from the device pointed to by handle
 *
 * calls:	SAir_ReadBARSize(), SAir_WriteBARSize(),
 *		pciCfgWriteShort(), pciCfgWriteByte(),
 *
 * returns:	void
 */

void SAir_AssignResources(U32 handle, U32 *sizePtr)
{
U32	data, offset = PCI_MEM_BAR ;
S32	i, status, barFlag = 0 ;

	/* Disable PCI IO and memory accesses */
	SAr_PciCfgWriteShort(handle, PCI_COMMAND, 0, &status) ;

	/* Scan each of the BARS for this device */
	for (i = 0; i < PCI_MAX_BAR; i++) {
		data = SAir_ReadBARSize(handle, sizePtr, offset) ;
		if (data != 0) {
			/* Assign some PCI memory space to this BAR */
			barFlag |= SAir_WriteBARSize(handle, data,
					*sizePtr, offset) ;
		}
		/* increment the BAR offset value */
		offset += 4 ;
		sizePtr++ ;
	}
	
	/* Check out the Expansion ROM, but don't give it any PCI space */
	sizePtr++ ;
	offset = PCI_ROM_BAR ;
	SAir_ReadBARSize(handle, sizePtr, offset) ;

	/* Enable PCI IO and memory accesses as found by scan, with MASTER */
	barFlag |= PCI_MASTER_ENABLE ;
	SAr_PciCfgWriteShort(handle, PCI_COMMAND, barFlag, &status) ;
	SAr_PciCfgWriteByte(handle, PCI_LATENCY, 0xFF, &status) ;
}

/*============================================================================
 *
 * routine:	SAr_ConfigurePciDevices(void)
 *
 * parameters:	void
 *
 * description:	this routine will scan the devNode tree looking for
 *		valid PCI devices and assign resources to each one.
 *
 * calls:	SAir_AssignResources()
 *
 * returns:	void
 */

void SAr_ConfigurePciDevices(void)
{
extern devNodeType	*devNodes ;
devNodeType		*current ;

	for (current = devNodes ; current != NULL; current = current->next) {
		/* check if device is PCI, if so it needs system resources */
		if ((current->handle) & PCI_TYPE) {
			/* call the resource manager to assign resources */
			SAir_AssignResources(current->handle, current->memSize) ;
		}
	}
}

/*============================================================================
 *
 * routine:	SAr_Mem2Pci(void *memAddress, U32 *status)
 *
 * parameters:	memAddress - StrongARM address to be converted to PCI space
 *		*status - flag for good/bad address.
 *
 * description: this routine will convert the given StrongARM address to PCI
 *		space. First, need to look up the PCI base address and amount
 *		of memory allocated to PCI - don't use constants because the
 *		devNode init code just slots the EBSA-285 into the next
 *		available PCI space.
 *
 * calls:	SAr_FindHostNode() - returns devNode for this device
 *
 * returns:	address in PCI memory space
 *		*status OK if address is in PCI space, else !OK
 */

void *SAr_Mem2Pci(void *memAddress, U32 *status)
{
devNodeType	*node ;
U32		 memPtr, pciBase, pciSize, bar ;

	memPtr = (U32)memAddress ;

	/* Simple mapping of StrongARM high memory -> PCI space */
	if (memPtr >= PCI_MEM) {
		*status = OK ;
		/* FIXME - target specific! */
		return (void *)(memPtr - PCI_MEM) ;
	}

	/* Else find our memory in PCI space:
	 * This target may have no memory mapped to PCI space.
	 */
	pciSize = PCI_DRAMSIZE ;
	/* Find our node */
	node = SAr_FindHostNode() ;

	if ((pciSize != 0) && (node != NULL)) {
		pciSize = node->memSize[HOST_MEMBAR] ;
		bar = HOST_MEMBAR ;
		pciBase = SAr_GetNextMemBAR(node->handle, &bar) ;

		/* Check the memory address is in range */
		if ((memPtr > PCI_DRAMBASE) &&
			(memPtr <= (PCI_DRAMBASE + pciSize))) {
			*status = OK ;
			return (void *)(pciBase + memPtr - (U32)PCI_DRAMBASE) ;
		}
	}

	*status = ~OK ;
	return (void *)0 ;
}

/*============================================================================
 *
 * routine:	SAr_GetNextMemBAR(U32 handle, U32 *bar)
 *
 * parameters:	handle - handle of device being accessed
 *		*bar - offset into Base Address Registers to start from
 *
 * description: this routine will scan the remaining Base Address Registers
 *		for the specified device. When a BAR is found which points to
 *		PCI memory space, the address in the BAR is returned.
 *
 * calls:	SAr_GetPciBAR()
 *
 * returns:	contents of BAR (0 if none found)
 *		*bar is modified to next possible BAR offset
 */

U32 SAr_GetNextMemBAR(U32 handle, U32 *bar)
{
U32	i, data ;
S32	status ;

	/* Scan remaining BARS for this device */
	for (i = *bar; i < PCI_MAX_BAR; i++) {
		/* Read back address assigned for this BAR */
		data = SAr_GetPciBAR(handle, i, &status) ;

		/* If not I/O, then it is memory */
		if ((data & PCI_IO_ENABLE) == 0) {
			/* Got one, set next available BAR & return address */
			*bar = i + 1 ;
			return (data & ~0x0F) ;
		}
	}

	/* Nothing found */
	*bar = 0 ;
	return 0 ;
}

/*============================================================================
 *
 * routine:	SAr_GetNextIOBAR(U32 handle, U32 *bar)
 *
 * parameters:	handle - handle of device being accessed
 *		*bar - offset into Base Address Registers to start from
 *
 * description: this routine will scan the remaining Base Address Registers
 *		for the specified device. When a BAR is found which points to
 *		PCI I/O space, the address in the BAR is returned.
 *
 * calls:	SAr_GetPciBAR()
 *
 * returns:	contents of BAR (0 if none found)
 *		*bar is modified to next possible BAR offset
 */

U32 SAr_GetNextIOBAR(U32 handle, U32 *bar)
{
U32	i, data ;
S32	status ;

	/* Scan remaining BARS for this device */
	for (i = *bar; i < PCI_MAX_BAR; i++) {
		/* Read back address assigned for this BAR */
		data = SAr_GetPciBAR(handle, i, &status) ;

		/* Check if it is I/O */
		if (data & PCI_IO_ENABLE) {
			/* Got one, set next available BAR & return address */
			*bar = i + 1 ;
			return (data & ~0x0F) ;
		}
	}

	/* Nothing found */
	*bar = 0 ;
	return 0 ;
}

/*============================================================================
 *
 * routine:	SAr_FindPCIDevClass(U32 devClass, U32 *inst)
 *
 * parameters:	devClass - device class to scan for
 *		*inst  - pointer to instance of card to find.
 *
 * description:	this routine will scan the devNode tree looking for devices
 *		which match the class code required. When a device matches
 *		the class code, the instance is checked - this enables
 *		software to uniquely address a device, even when more than
 *		one may be fitted.
 *
 * calls:	SAr_GetPciClassID()
 *
 * returns:	handle to required device (0 if not found)
 *		*inst is modified to next possible instance
 */

U32 SAr_FindPCIDevClass(U32 devClass, U32 *inst)
{
extern devNodeType	*devNodes ;
devNodeType		*current ;
U32			 newClass, count = 0 ;
S32			 status ;

	for (current = devNodes ; current != NULL; current = current->next) {
		/* check if device is PCI, if so check the device class */
		if ((current->handle) & PCI_TYPE) {
			newClass = SAr_GetPciClassID(current->handle, &status) ;

			if (newClass == devClass) {
				if (count++ >= *inst) {
					*inst = count ;
					return (current->handle) ;
				}
			}
		}
	}

	/* Nothing found */
	*inst = 0 ;
	return 0;
}

/*============================================================================
 *
 * routine:	SAr_FindPCIDevice(U16 vendor, U16 device, U32 *inst)
 *
 * parameters:	vendor - vendor code to scan for
 *		device - device code to scan for
 *		*inst  - pointer to instance of card to find.
 *
 * description:	this routine will scan the devNode tree looking for devices
 *		which match the vendor & device code required. When a device
 *		matches both vendor and device codes, the instance is checked
 *		- this enables software to uniquely address a device, even
 *		when more than one may be fitted. If only a specific vendor
 *		or device code is required (not both), the other code can
 *		be ignored by specifying PCI_NOTFITTED.
 *
 * calls:	SAr_GetPciVendorID(), SAr_GetPciDeviceID()
 *
 * returns:	handle to required device (0 if not found)
 *		*inst is modified to next possible instance
 */

U32 SAr_FindPCIDevice(U16 vendor, U16 device, U32 *inst)
{
extern devNodeType	*devNodes ;
devNodeType		*current ;
U32			 value, count = 0 ;
S32			 status ;

	for (current = devNodes ; current != NULL; current = current->next) {
		/* check if device is PCI, if so check the device class */
		if ((current->handle) & PCI_TYPE) {
			value = SAr_GetPciVendorID(current->handle, &status) ;

			/* Allow don't care vendor code or match */
			if ((vendor == PCI_NOTFITTED) || (value == vendor)) {
				value = SAr_GetPciDeviceID(current->handle, &status) ;

				/* Allow don't care device code or match */
				if ((device == PCI_NOTFITTED) || (value == device)) {
					if (count++ >= *inst) {
						*inst = count ;
						return (current->handle) ;
					}
				}
			}
		}
	}

	/* Nothing found */
	*inst = 0 ;
	return 0;
}

#endif

