/* -*-C-*-
 *
 * Copyright (c) 1996 Advanced RISC Machines Limited
 * All Rights Reserved
 *
 * Changed for use with different stack
 * Geoff Smith - Copyright Digital Semiconductor 
 * Spring 1997
 *
 * ethernet.c:  Angel drivers for Ethernet using UDP/IP stack
 */


#include <string.h>
#include "devconf.h"

#define in_sockaddr sockaddr_in /* Use BSD like definition */

#include "address.h"
#include "udp.h"
#include "angel.h"
#include "channels.h"           /* for getting buffers */
#include "endian.h"             /* PUT16LE */
#include "logging.h"            /* __rt_warning() */
#include "support.h"            /* __rt_[en/dis]ableIRQ() */
#include "serlock.h"            /* serialisation, etc. */
#include "params.h"             /* parameter structures and utilities */
#include "arm_ether.h"
#include "socket.h"
#include "debugio.h"
#include "ethernet.h"
#include "errors.h"

#if (ETHERNET_SUPPORTED == 1)


#ifdef DEBUG
static int ethernet_debug = 1;
#endif
extern int InitialisationFail;

/*
 * forward declarations of necesary routines
 */
static DevError ethernet_Write(DeviceID devid, p_Buffer buffer,
                               unsigned int length, DevChanID devchan);

static DevError ethernet_RegisterRead(DeviceID devid,
                                      DevChanID devchan);

static DevError ethernet_Control(DeviceID devid, DeviceControl op, void *arg);

const struct angel_DeviceEntry angel_EthernetDevice =
{
    DT_ANGEL,
    {
        ethernet_Write, ethernet_RegisterRead
    },
    ethernet_Control,
    0,
    {
        0, NULL
    },
    {
        0, NULL
    }
};

/*
 * the three sockets that we talk down
 */
static int ctrlsock = -1, dbugsock = -1, applsock = -1;

/*
 * holder for errors
 */
static int errno;
static int rx_flow = 1; /* see do_poll()/ethernet_init() */

/*
 * The host we are currently talking to.  Assumed to be the last host
 * that we received anything from
 */
 
static struct in_sockaddr host;
static int hostlen=0;

/* Our client IP address */
#if (USE_RARP == 1)	|| (USE_BOOTP == 1)
unsigned char  client_ipaddr[4] = {0,0,0,0}; /* Fusion used hardwired address,
										This stack uses RARP, BOOTP, or Hardwired address
										See USE_RARP in devconf.h  (The IP addres is also
										in devconf.h) */

#else
unsigned char client_ipaddr[4] = {__IP1, __IP2, __IP3, __IP4 };
#endif
				   

/**********************************************************************/
/*
 *  Function: open_socket
 *   Purpose: Open a non-blocking UDP socket, and bind it to a port number
 *
 *    Params:
 *       Input: port    The port to bind to
 *
 *   Returns:
 *          OK: Socket number
 *       Error: -1
 */

static int open_socket(const unsigned int port)
{
    int sock;
    struct in_sockaddr to;

    /*
     * open the socket, and try to bind it to the given port
     */
    if (0 > (sock = socket(AF_INET, SOCK_DGRAM, 0)))
    {
        return -1;
    }

    /*
     * try to bind to the port
     */
    to.sin_family = AF_INET;
    to.sin_port = htons(port);    
	if ( port > 0 )
       to.sin_addr.s_addr = INADDR_ANY;
    else
    {
        __rt_memcpy( &to.sin_addr.s_addr, client_ipaddr,
                     sizeof(to.sin_addr.s_addr) );
    }
    if (0 != (errno = bind(sock, &to, sizeof(to))))
    {
        close(sock);
        return -1;
    }

    /*
     * mark this socket as non-blocking
	 * 
     */
    if (0 != (errno = nonblocking(sock)))
    {
        close(sock);
        return -1;
    }

    /*
     * socket is ready for action
     */
    return sock;
}

/*
 *  Function: findport
 *   Purpose: Ascertain the port number bound to a socket
 *
 *    Params:
 *       Input: sock    socket descriptor
 *
 *   Returns:
 *          OK: port (in network byte order)
 *       Error: 0
 */
static unsigned short int findport(int sock)
{
    struct in_sockaddr name;
    int namelen;

    namelen = sizeof(name);
    if (getsockname(sock, (struct sockaddr *)&name, &namelen) != 0)
    {
        return 0;
    }
    else
        return name.sin_port;
}

/*
 *  Function: do_ctrlpoll
 *   Purpose: Poll the control socket for received packets
 *
 *    Params: None
 *
 *   Returns: Nothing
 */
static void do_ctrlpoll(void)
{
    static const char ctrlmagic[] = CTRL_MAGIC;
    CtrlResponse dialogue;
    struct in_sockaddr remote;
    int remotelen = sizeof(remote);

    /*
     * see whether anything is available on the control socket
     */

	 if (0 > (errno = recvfrom(ctrlsock, (char *)dialogue, sizeof(dialogue), 0,
                 (struct sockaddr *)&remote, &remotelen)))
    {
    }
    else if (__rt_strcmp(ctrlmagic, (char *)dialogue) == 0)
    {
        /*
         * this is a valid control packet - send the port numbers
         * back to the requester
         */
        unsigned short *sptr = (unsigned short *)(dialogue + RESP_DBUG);

        *sptr   = findport(dbugsock);
        *++sptr = findport(applsock);

        if ( 0 > (errno = sendto(ctrlsock, (char *)&dialogue, sizeof(dialogue), 0,
                   (struct sockaddr *)&remote, remotelen)))
	  {
	  }
        else
        {
            /*
             * make a copy of the remote address of the client
             */
            __rt_memcpy(&host, &remote, remotelen);
            hostlen = remotelen;
        }
    }

}

/*
 *  Function: do_poll
 *   Purpose: Poll an open socket for received packets
 *
 *    Params:
 *       Input: sock    The socket to poll
 *
 *              devchan The device channel associated with the socket
 *
 *   Returns: Nothing
 */
 
#define UDP_RQ_MAX 1526 /* Don't know exactly what ARM set this to 
					- Guessing around this value as its the largest size allowed
					across an ethernet, 1526, this includes the IP/UDP headers */

static void do_poll(const int sock, DevChanID devchan)
{
    int nbytes;
    p_Buffer buffer;
    static char frmbuf[UDP_RQ_MAX];

    /* if we've been disabled from returning packets, don't ask for them! */
    if (rx_flow == 0)
      {
      LogWarning(LOG_ETHER, ("do_poll: rx_flow == 0\n", errno));
      return;
      }

    /*
     * Unfortunately, we don't know how long the next packet is going
     * to be, so we have a private MTU buffer, and copy into an
     * allocated buffer when we know how much space to ask for.  If
     * this proves to be too inefficient, then we will have to look
     * at other ways of doing this.
     */
	
    if ((nbytes = recv(sock, frmbuf, UDP_RQ_MAX, 0)) < 0) 
      {
	
	/* return; */		
      }
    else if( nbytes != 0 )
      {
	if (rx_flow < 0)
	  rx_flow ++;

	/*
	 * pass this along to higher levels
	 */
	if ((buffer = angel_DD_GetBuffer(DI_ETHERNET,devchan,nbytes)) != NULL)
	  {
	    __rt_memcpy(buffer, frmbuf, nbytes);
	    angel_DD_GotPacket(DI_ETHERNET, buffer, nbytes, DS_DONE, devchan);
	  }
	else
	  LogWarning(LOG_ETHER, ("do_poll: couldn't DD_GetBuffer"));
      }


}

/*
 *  Function: ethernet_init
 *   Purpose: Initialise the UDP stack
 *
 *    Params: Nothing
 *
 *   Returns:
 *          OK: DE_OKAY
 *       Error: DE_NO_DEV
 */
static DevError ethernet_init(void)
{
    /* rx_flow == 1 enabled packet flow; == 0 disables it, < 0 specifies the
       number of packets allowed through. See DC_RX_PACKET_FLOW && serpkt.c */
    rx_flow = 1;

  /*
   * try to initialise the three sockets
   */
  if ((ctrlsock = open_socket(CTRL_PORT)) < 0)
    {
      return DE_NO_DEV;
    }
  
  if ((dbugsock = open_socket(0)) < 0)
    {
      (void)close(ctrlsock);
      ctrlsock = -1;
      
     return DE_NO_DEV;
    }
  
  if ((applsock = open_socket(0)) < 0)
    {
      (void)close(ctrlsock);
      ctrlsock = -1;

      (void)close(dbugsock);
      dbugsock = -1;

      return DE_NO_DEV;
    }

  /*
   * it worked! (Don't be so surprised)
   */
  return DE_OKAY;
}

/**********************************************************************/

/*
 *  Function: ethernet_Write
 *   Purpose: Entry point for asynchronous writes to the ethernet device.
 *            See documentation for angel_DeviceWrite in devdriv.h
 */
static DevError ethernet_Write(DeviceID devid, p_Buffer buffer,
                               unsigned int length, DevChanID devchan)
{
    int sock, nbytes;
    DevStatus status = DS_DONE;

    IGNORE(devid);


	/*
		Angel does not know the device is not there
		so lie at this point.
	*/
	if (InitialisationFail)
		return DE_OKAY;
		
    if (devchan == DC_DBUG)
        sock = dbugsock;
    else if (devchan == DC_APPL)
        sock = applsock;
    else
    {
        return DE_BAD_CHAN;
    }

    /*
     * we need to broadcast the first message sent (the booted,
     * message), but talk point-to-point once an host has been
     * identified
	 */
    if (hostlen == 0)
    {
        unsigned long bcastaddr = htonl((unsigned int)IP_BROADCAST);

        host.sin_family = AF_INET;
        host.sin_addr.s_addr = bcastaddr;
        hostlen = sizeof(host);
    }
	
    Angel_EnterSVC();
	if (0 > (nbytes = sendto(sock, (char *)buffer, length, 0,(struct sockaddr *)&host, hostlen)))
    {
        status = DS_DEV_ERROR;
	    Angel_ExitToUSR();		
		return DE_FAILED;
    }

    /*
     * schedule a callback to say the packet has gone 
	 * But why would I want a callback if I've sent the packet???
	 * Because the callback is to the channel that originated the write.
     */
    angel_DD_SentPacket(DI_ETHERNET, buffer, length, status, devchan);

    /* now it's okay to resume in user mode */

	Angel_ExitToUSR();
    return DE_OKAY;
}

/*
 *  Function: ethernet_RegisterRead
 *   Purpose: Entry point for asynchronous reads from the ethernet device.
 *            See documentation for angel_RegisterRead in devdriv.h
 */
static DevError ethernet_RegisterRead(DeviceID devid, DevChanID devchan )
{
    IGNORE(devid);
    IGNORE(devchan);
    /* nothing to do */
    return DE_OKAY;
}

/*
 *  Function: ethernet_Control
 *   Purpose: Entry point for ethernet device control functions.
 *            See documentation for angel_DeviceControl in devdriv.h
 */
static DevError ethernet_Control(DeviceID devid, DeviceControl op, void *arg)
{
    DevError ret_code;

    IGNORE(devid);
    IGNORE(arg);

    Angel_EnterSVC();

    /* we only implement the standard ones here */
    switch (op)
    {
    case DC_RX_PACKET_FLOW:
      LogInfo(LOG_ETHER, ( "ethernet_Control: set rx_flow to %d (was %d)\n",
                                 (int) arg, rx_flow));
      rx_flow = (int)arg;
      ret_code = DE_OKAY;
      break;

    case DC_INIT:
	  		/* Angel ignores this return code
			   so it never knows the device failed.
			   Hence must make this layer lie to Angel instead
			   by saying in has no data waiting rather than errors
			   ditto for writes.
			*/
			
      if (netboot_main()) 
	ret_code = DE_NO_DEV;
      else				
	ret_code = ethernet_init();
				
      break;

#if 0
      case DC_RESET:         
          ret_code = ethernet_reset(devid, port);
          break;

      case DC_RECEIVE_MODE:  
          ret_code = ethernet_recv_mode(devid, (DevRecvMode)((int)arg));
          break;

      case DC_SET_PARAMS:     
          ret_code = ethernet_set_params((const ParameterConfig *const)arg);
          break;
#endif

      default:
          ret_code = DE_BAD_OP;
          break;
    }

    Angel_ExitToUSR();

	if (ret_code != DE_OKAY) {
		angel_ChannelSelectDevice(DI_SERIAL);
	}
	
    return ret_code;
}

/**********************************************************************/

/*
 *  Function: angel_EthernetPoll
 *   Purpose: Poll for newly arrived packets.  See function prototype in
 *              arm_ether.h for a full description.
 *	If the active device changes from ethernet then don't check the 
 *	stack anymore - the user has opted for serial configuration.
 *
 */
void angel_EthernetPoll(unsigned int data)
{
DeviceID active_device;
	
	if (InitialisationFail)
		return;

	angel_ChannelReadActiveDevice(&active_device);	
	if (active_device != DI_ETHERNET) {
		return;
	}


    Angel_EnterSVC();
    IGNORE(data);
    /*
     * poll control device...
     */
    do_ctrlpoll();

    /*
     * ... then poll both channel devices
     */
    do_poll(dbugsock, DC_DBUG);
    do_poll(applsock, DC_APPL);

    Angel_ExitToUSR();

}


void angel_EthernetNOP(unsigned int data)
{
    IGNORE(data);
}


/* EOF ethernet.c */
#endif
