/****************************************************************************
**                                                                          *
**  Copyright  1997						     								*
**  by Digital Equipment Corporation, Maynard, Massachusetts.		     	*
**                                                                          *
**  All Rights Reserved							     						*
**                                                                          *
**  Permission  is  hereby  granted  to  use, copy, modify and distribute   *
**  this  software  and  its  documentation,  in  both  source  code  and   *
**  object  code  form,  and without fee, for the purpose of distribution   *
**  of this software  or  modifications  of this software within products   *
**  incorporating  an  integrated   circuit  implementing  Digital's  		*
**  StrongARM architecture,  regardless  of the  source of such integrated 	*
**	circuit, provided that the  above copyright  notice and this permission *
**  notice appear  in  all copies,  and  that  the  name  of  Digital  		*
**  Equipment Corporation  not  be  used  in advertising or publicity 		*
**  pertaining to distribution of the  document  or  software without 		*
**  specific, written prior permission.							            *
**                                                                          *
**  Digital  Equipment  Corporation   disclaims  all   warranties  and/or   *
**  guarantees  with  regard  to  this  software,  including  all implied   *
**  warranties of fitness for  a  particular purpose and merchantability,   *
**  and makes  no  representations  regarding  the use of, or the results   *
**  of the use of, the software and documentation in terms of correctness,  *
**  accuracy,  reliability,  currentness  or  otherwise;  and you rely on   *
**  the software, documentation and results solely at your own risk.	    *
**                                                                          *
**  StrongARM is a trademark of Digital Equipment Corporation               *
**                                                                          *
*****************************************************************************/
#ifndef LINT
extern char *rcsid;
#endif

/*
 * DESCRIPTION:
 *     This module implements the ARP protocol as described in RFC 826. Only
 *     MAC style hardware addresses and IP style protocol addresses are
 *     supported. In addition this module allows other modules to query 
 *     the information received through the ARP protocol.
 *
 *
 */

#include "net_buff.h"
#include "ether.h"
#include "netparam.h"
#include "arp.h"
#include "ethernet.h"
#include "ip.h"
#include "support.h"
#include "debugio.h"
#include "errors.h"
#include "devconf.h"
#include "timer.h"


#if (ETHERNET_SUPPORTED == 1)

/*
 * Protocol structures and field values
 * ====================================
 */

/* Arp protocol fields. Defined using offset constants etc. to avoid any
 * allignment or byte ordering problems when using structures. All offsets
 * are from the start of the ARP packet (not from the start of the ethernet
 * frame) Note that  since this implementation can only be used for resolving 
 * IP addresses over LANs all address sizes (hardware and protocol) are fixed.
 *
 */
#define AR_HRD_OFFSET 0                  /* Hardware address space - must be Ethernet address space */
#define AR_PRO_OFFSET (AR_HRD_OFFSET+2)  /* Protocol address space - must be IP; value is ethernet protocol id
					   					 of IP (0x0800) */
#define AR_HLN_OFFSET (AR_PRO_OFFSET+2)  /* Byte length of each hardware address; will be 6 */
#define AR_PLN_OFFSET (AR_HLN_OFFSET+1)  /* Byte length of each protocol address; will be 4 */
#define AR_OP_OFFSET  (AR_PLN_OFFSET+1)   /* opcode */
#define AR_SHA_OFFSET (AR_OP_OFFSET +2)  /* Source hardware address */
#define AR_SPA_OFFSET (AR_SHA_OFFSET+MAC_ADDRESS_SIZE)  /* Source IP address */
#define AR_THA_OFFSET (AR_SPA_OFFSET+IP_ADDRESS_SIZE)   /* Target hardware address (if known) */
#define AR_TPA_OFFSET (AR_THA_OFFSET+MAC_ADDRESS_SIZE)  /* Target protocol address */
#define AR_PACKET_SIZE (AR_TPA_OFFSET+IP_ADDRESS_SIZE)  /* Total size of arp packets */

/*
 * Constants used in the ARP protocol 
 */

/* Op codes; defined in RFC 826 */
static char const ar_op_request[2] = {0x00, 0x01};
static char const ar_op_reply[2] = {0x00, 0x02};

#if (USE_RARP == 1)
static char const ar_op_rarp_reply[2] = {0x00, 0x04};
#endif

/* Ethernet hardware type - see RFC 1340 */
char const arp_hardware_ether[2] = {0x00, 0x01};

/*
 * Module data structures
 * ======================
 *
 */ 

/*
 * ARP table.  Contains ethernet<->ip address pairs. There is a copy for
 * each ethernet device.
 */
#define ARP_TABLE_SIZE 5

static struct arp_data {  
  ip_addr     ip_address;
  mac_addr    enet_addr;
  enum        {isempty, isvalid} flags;
} arp_table[MAX_ETHER_DEVICES][ARP_TABLE_SIZE];

/* List of local ip addresses; assumed to be one per ethernet device,
 * note that static data initialisation will initialise all of these
 * to 0 which in ip means "I don't know my own address"
 */
static ip_addr device_ip_address[MAX_ETHER_DEVICES];
extern char client_ipaddr[]; /* Its in arm_ether.c */
/*
 * Internal functions
 * ==================
 */

/* arp_lookup - internal function to search for an ip address in the table
 *
 * Argument:
 *    device_no - device number on which we are planning to send
 *    ip        - ip address to search for
 *
 * Returned value:
 *    +ve or 0 - Index into table of entry containing value
 *    -1       - Not found
 */
static int arp_lookup(int const device_no, ip_addr const ip)
{
  int i,j, failed;

	for (j=0;j<ARP_TABLE_SIZE;j++){
		if (arp_table[device_no][j].flags==isvalid) {
			failed = FALSE;
			for (i=0;i<IP_ADDRESS_SIZE;i++) {
				if (arp_table[device_no][j].ip_address[i] != ip[i])
					failed = TRUE;
			}			
			if (failed == FALSE)
	    		return j;
		}	
  	}

  return ERR_ARPLOOKUPFAIL;
}

/*
 * 
 * arp_query_network - internal function to ask whether anybody on the network
 *                     recognises an address. 
 *
 * Arguments:
 *     device_no - which device to query
 *     ip        - ip address required
 */
static void arp_query_network(int const device_no,ip_addr const ip)
{
    /* Broadcast an arp query message.  Then read the input stream until
     * an arp message is received or 5 seconds elapses.
     */
    
    /* Declare an initial buffer and a static variable in which a record of the current buffer can
     * be maintained 
     */

    mac_addr my_mac_address;
	unsigned char *frame_buffer, *balance_buffer;
    unsigned char* arp_packet;
    int i, err;

	if (NULL == (frame_buffer = framemalloc(ETHER_BUFFER_SIZE))) {
		return;
	}
	arp_packet = (frame_buffer + ethernet_data_offset());
    /*
     * Set up a request packet
     */
    *(unsigned char *)(arp_packet+AR_HRD_OFFSET)=arp_hardware_ether[0];
    *(unsigned char *)(arp_packet+AR_HRD_OFFSET+1)=arp_hardware_ether[1];
    for(i=0; i<ETHERNET_PROTOCOL_ID_SIZE; i++)
		*(unsigned char *)(arp_packet+AR_PRO_OFFSET+i)=ether_protocol_ip[i];
    *(unsigned char *)(arp_packet+AR_HLN_OFFSET)= MAC_ADDRESS_SIZE;
    *(unsigned char *)(arp_packet+AR_PLN_OFFSET)= IP_ADDRESS_SIZE;
    *(unsigned char *)(arp_packet+AR_OP_OFFSET)=ar_op_request[0];
    *(unsigned char *)(arp_packet+AR_OP_OFFSET + 1)= ar_op_request[1];
    
    /* Get the device's MAC address and copy it in */
    ether_device_get_hw_address(device_no, my_mac_address);
    for(i=0; i<MAC_ADDRESS_SIZE; i++)
		*(unsigned char *)(arp_packet+AR_SHA_OFFSET+i)=my_mac_address[i];

    /* Copy in the device's IP address */
    for(i=0; i<IP_ADDRESS_SIZE; i++)
		*(unsigned char *)(arp_packet+AR_SPA_OFFSET+i)=device_ip_address[device_no][i];

    /* And the requested IP address */
    for(i=0; i<IP_ADDRESS_SIZE; i++)
		*(unsigned char *)(arp_packet+AR_TPA_OFFSET+i)=ip[i];
    /*
     * Transmit the packet.
     */
	err = ethernet_write(device_no,
			    arp_packet,
			    broadcast_mac_address,
			    ether_protocol_arp,
			    AR_PACKET_SIZE,
			    frame_buffer,
			    &balance_buffer);

/*
	 At this point it would be preferable to use a timeout and check for/send
	 ARP info for a few seconds - however I don't want to take up a timer for
	 something that will so rarely be used in most setups, ie only one host.
	 It is here for those who will be writing/extending this code for 'proper' ARPing.
	 Geoff Smith, Summer 1997.
*/

//    while( (difftime(time(0),tod) < 5) && arp_lookup(device_no, ip)<0)
    ethernet_process_one_packet(device_no, balance_buffer, &balance_buffer);
	framefree(balance_buffer);


}

/* arp_input - arp received packet processor.
 *
 * Arguments:
 *     protocol_data - protocol data passed to module when protocol registered; not used
 *     device_no     - device on which packet was received;
 *     packet        - packet, with ethernet header removed;
 *     size          - size of packet;
 *     source        - source MAC address of packet;
 *     frame_buffer  - original buffer containing packet.
 *     balance_buffer- buffer returned to maintain buffer balance buffer.
 *
 * Description:
 *     Processes a received ARP packet. For a detailed explanation of the 
 *     interface see ethernet.h.
 *
 */
static void arp_input(void * protocol_data,
		      int device_no,
		      unsigned char* packet,
		      int size,
		      mac_addr source,
		      unsigned char* frame_buffer,
		      unsigned char* * balance_buffer)
{
    int merge_flag = FALSE;
	
    int tidx;
    mac_addr my_mac_address;
    int i;
	static int victim_slot = 1;


	*balance_buffer=frame_buffer;
	/*
		Is it a RARP reply?
	*/
#if (USE_RARP == 1)
	if (0 == __rt_memcmp(ar_op_rarp_reply, (unsigned char*)(packet + AR_OP_OFFSET), 2 )) {
		rarp_input(protocol_data, device_no, packet, size,
				source, frame_buffer, balance_buffer);
		return;		
	}
#endif	
	
	
    /*
     * Quit if the hardware type isn't ethernet or if the protocol
     * type isn't IP.
     */
    if(__rt_memcmp(arp_hardware_ether, packet+AR_HRD_OFFSET, 2) != 0 ||
       __rt_memcmp(ether_protocol_ip, packet+AR_PRO_OFFSET,
		       ETHERNET_PROTOCOL_ID_SIZE) != 0) {
		return;
    }

    /*
     * Next check whether we already know about the IP source address
     */

    for (tidx=0;tidx<ARP_TABLE_SIZE;tidx++) {
		if ((arp_table[device_no][tidx].flags==isvalid) &&
		    (__rt_memcmp(arp_table[device_no][tidx].ip_address, 
				     packet+AR_SPA_OFFSET, IP_ADDRESS_SIZE)==0)){

		    /*  An entry already exists for this address; update it. */
		    for(i=0; i<MAC_ADDRESS_SIZE; i++)
			arp_table[device_no][tidx].enet_addr[i] = 
			  *(unsigned char *)( packet+AR_SHA_OFFSET+i);
		    merge_flag = TRUE;
		    break;
		}
    }
	
    /* Nothing more to do unless we are the target */

    if(__rt_memcmp(device_ip_address[device_no], packet+AR_TPA_OFFSET, 
		       IP_ADDRESS_SIZE) != 0) {
		return;
    }

    /*
     * If we are the target, continue processing.
     */
    if (!merge_flag){
		/*
		 * No entry from the sender is already in the table.
		 * Insert the sender protocol and hardware address pair in the
		 * table.
		 * If we don't find an empty slot, pick a victim slot at random(ish).
		 * Victim will not be slot 0 - this is used by higher
		 * layers to prevent overwriting of the Angel host address
		 * Enter the sender data in the table.
		 */
		for (tidx=0;tidx<ARP_TABLE_SIZE;tidx++)
			if (arp_table[device_no][tidx].flags==isempty)
		    	break;
		
		if (tidx >= ARP_TABLE_SIZE) { /* Pick a slot then */
			tidx = victim_slot++;
			victim_slot = (victim_slot % (ARP_TABLE_SIZE-1)) + 1;
		}

		for(i=0; i<IP_ADDRESS_SIZE; i++) {
		    arp_table[device_no][tidx].ip_address[i] = 
		      *(unsigned char *)( packet+AR_SPA_OFFSET+i);
		}

		for(i=0; i<MAC_ADDRESS_SIZE; i++) {
		    arp_table[device_no][tidx].enet_addr[i] = 
		      *(unsigned char *)( packet+AR_SHA_OFFSET+i);
		}
		arp_table[device_no][tidx].flags=isvalid;
    }

    /*
     * If the packet is a request packet we must send a response
     * packet.
     */
    if (__rt_memcmp(ar_op_request, packet+AR_OP_OFFSET, 2)== 0){
		mac_addr mac;
		int err;
		/*
		 * Move sender data to target data.
		 * Put our address into the sender fields.
		 * Set the opcode to reply.
		 * Transmit the packet.
		 */

		/* Copy target hardware and protocol addresses from the original source addresses */
		for(i=0; i<MAC_ADDRESS_SIZE; i++) {
		    *(unsigned char *)(packet + AR_THA_OFFSET+i)=
					  *(unsigned char *)(packet + AR_SHA_OFFSET+i);
		    mac[i] = *(unsigned char *)(packet + AR_SHA_OFFSET+i);
		}
		for(i=0; i<IP_ADDRESS_SIZE; i++)
		    *(unsigned char *)(packet + AR_TPA_OFFSET+i)=
					  *(unsigned char *)(packet + AR_SPA_OFFSET+i);

		/* The new source hardware address is the device's hardware address */
		ether_device_get_hw_address(device_no, my_mac_address);
		for(i=0; i<MAC_ADDRESS_SIZE; i++)
		    *(unsigned char *)(packet + AR_SHA_OFFSET+i)=my_mac_address[i];

		/* The new source protocol address is the device's protocol address */
		for(i=0; i<IP_ADDRESS_SIZE; i++)
		    *(unsigned char *)(packet + AR_SPA_OFFSET+i)=device_ip_address[device_no][i];

		/* The new op code is reply */
		*(unsigned char *)(packet + AR_OP_OFFSET)= ar_op_reply[0];
		*(unsigned char *)(packet + AR_OP_OFFSET +1)= ar_op_reply[1];

		/* All other fields are identical to the received packet; send the packet */

		err = ethernet_write(device_no,
			       packet,
			       mac,
			       ether_protocol_arp,
			       AR_PACKET_SIZE,
			       frame_buffer,
			       balance_buffer);
    }	
}

/*
 * Externally callable functions
 * =============================
 */


/* arp_init - initialises the arp module; and registers it with the
 * ethernet protocol handler.
 *
 * Argument:
 *    device_no - device number on which arp should be registered.
 */
void arp_init(int device_no)
{
    int entry_no;
    
    /* Clear out all tables */
    __rt_memcpy(device_ip_address[device_no], client_ipaddr, 
	   IP_ADDRESS_SIZE);
	

    for(entry_no=0; entry_no < ARP_TABLE_SIZE; entry_no++) {
	arp_table[device_no][entry_no].flags = isempty;
    }

    /* Register the ARP protocol with the ethernet protocol handler */
    ethernet_register_protocol(device_no, ether_protocol_arp, 
	arp_input, NULL);

}
	

/*
	arp_force_host_address - (int dev_no, ip_addr ip, mac_addr mac)
	sets table location 0 to ip/mac pair - this will not be
	overwritten by ARP packets.
	(This was added for the Angel stuff cos it only talks to one host and I want to
	protect that address combination)
*/
int arp_force_host_address(int device_no, ip_addr ip, mac_addr mac)
{
	int i;
	
   	if (device_no > MAX_ETHER_DEVICES) {
		return ERR_ARPUNKNOWNDEVICE;
	}
	
	for(i=0; i<IP_ADDRESS_SIZE; i++)
	    arp_table[device_no][0].ip_address[i] = *(unsigned char *)(ip+i);
	     
	for(i=0; i<MAC_ADDRESS_SIZE; i++)
	    arp_table[device_no][0].enet_addr[i] = *(unsigned char *)(mac + i);
	     
	arp_table[device_no][0].flags=isvalid;

return 0;	
}

/* arp_set_device_addr - tells arp a device's ip address.
 *
 * Arguments:
 *     device_no - device number
 *     ip        - ip address
 */
void arp_set_device_addr(int device_no,ip_addr ip)
{
    __rt_memcpy(device_ip_address[device_no], ip, IP_ADDRESS_SIZE);
}


/*
	arp_add - adds an ip/mac addres pair to the lookup table
	used in the IP layer
*/
void arp_add(int device_no, ip_addr ip, mac_addr mac)
{
	int idx;
	static unsigned int victim=1;
	

	idx = arp_lookup(device_no, ip);
	if (-1 == idx) {
		idx = (victim++ % (ARP_TABLE_SIZE-1)) + 1;  /* Kill an innocent bystander. Probably deserves it anyway. */	
		__rt_memcpy(arp_table[device_no][idx].ip_address, ip,IP_ADDRESS_SIZE);		
	}
	__rt_memcpy(arp_table[device_no][idx].enet_addr, mac, MAC_ADDRESS_SIZE);
	arp_table[device_no][idx].flags = isvalid;
	
}
	

 
/* arp_resolve - converts an ip address into an ethernet address
 *
 * Arguments:
 *     device_no - device on which the message is to be sent
 *     ip        - ip address
 *     mac       - returned ethernet MAC address
 *
 * Returned value:
 *     0      - address resolved; eaddr valid
 *     -VE     - Unable to resolve address.
 */
int arp_resolve(int device_no, ip_addr ip, mac_addr mac)
{
    int i,j;
    /*
     * If the ip address is broadcast, use broadcast ethernet address.
     * Look up the address in the table.  If found, return it.  If not
     * found, call arp_transact to get address from a server.
     * Repeat the loop 3 times.
     */

    if (__rt_memcmp(ip,local_ip_broadcast_address,IP_ADDRESS_SIZE) == 0){
		__rt_memset( (unsigned char *)mac ,(unsigned char)0xff ,6);
		return 0;
    }
    
    for (i=0;i<3;i++){
		j = arp_lookup(device_no, ip);
		if (j >= 0){ /* Found entry, okay */
	    	__rt_memcpy((char*) mac, (char*) arp_table[device_no][j].enet_addr, MAC_ADDRESS_SIZE);
	    	return 0;
		}
		
		arp_query_network(device_no, ip);
		/* Needs to wait for response */
		_msleep(500); // 1/2 second
    }
    // repeat the check a fourth time incase the final query worked 
    j = arp_lookup(device_no, ip);
    if (j>=0){
		__rt_memcpy((char*) mac, (char*) arp_table[device_no][j].enet_addr, MAC_ADDRESS_SIZE);
		return 0;
    }
    return ERR_ARPRESOLVEFAIL;
}

/* 
	arp_init_module - initialise the module 
 */
void arp_init_module(void)
{
/* Nothing to do! */
return;
}

/* 
 * arp_show - function to show all the entries in the arp table.
 *
 * Argument:
 *    None.
 *
 * Returned value:
 *    None.
 */
void arp_show()
{
return  ;
}

#endif
