/****************************************************************************
**                                                                          *
**  Copyright  1997						     								*
**  by Digital Equipment Corporation, Maynard, Massachusetts.		     	*
**                                                                          *
**  All Rights Reserved							     						*
**                                                                          *
**  Permission  is  hereby  granted  to  use, copy, modify and distribute   *
**  this  software  and  its  documentation,  in  both  source  code  and   *
**  object  code  form,  and without fee, for the purpose of distribution   *
**  of this software  or  modifications  of this software within products   *
**  incorporating  an  integrated   circuit  implementing  Digital's  		*
**  StrongARM architecture,  regardless  of the  source of such integrated 	*
**	circuit, provided that the  above copyright  notice and this permission *
**  notice appear  in  all copies,  and  that  the  name  of  Digital  		*
**  Equipment Corporation  not  be  used  in advertising or publicity 		*
**  pertaining to distribution of the  document  or  software without 		*
**  specific, written prior permission.							            *
**                                                                          *
**  Digital  Equipment  Corporation   disclaims  all   warranties  and/or   *
**  guarantees  with  regard  to  this  software,  including  all implied   *
**  warranties of fitness for  a  particular purpose and merchantability,   *
**  and makes  no  representations  regarding  the use of, or the results   *
**  of the use of, the software and documentation in terms of correctness,  *
**  accuracy,  reliability,  currentness  or  otherwise;  and you rely on   *
**  the software, documentation and results solely at your own risk.	    *
**                                                                          *
**  StrongARM is a trademark of Digital Equipment Corporation               *
**                                                                          *
*****************************************************************************/

#include "devconf.h"

#if (ETHERNET_SUPPORTED == 1)
#if (USE_BOOTP == 1)
 
#include "net_buff.h"
#include "netparam.h"
#include "ether.h"
#include "address.h"
#include "bootp.h"
#include "udp.h"
#include "ip.h"
#include "edevice.h"
#include "support.h"
#include "serial.h"
#include "platform.h"
#include "timer.h"
/*
 * Structure of a bootp packet (see rfc 951)
 */
#define BOOTPP_OP_OFFSET     0
#define BOOTPP_HTYPE_OFFSET  (BOOTPP_OP_OFFSET+1)
#define BOOTPP_HLEN_OFFSET   (BOOTPP_HTYPE_OFFSET+1)
#define BOOTPP_HOPS_OFFSET   (BOOTPP_HLEN_OFFSET+1)
#define BOOTPP_XID_OFFSET    (BOOTPP_HOPS_OFFSET+1)
#define BOOTPP_SECS_OFFSET   (BOOTPP_XID_OFFSET+4)
#define BOOTPP_CIADDR_OFFSET (BOOTPP_SECS_OFFSET+4)
#define BOOTPP_YIADDR_OFFSET (BOOTPP_CIADDR_OFFSET+4)
#define BOOTPP_SIADDR_OFFSET (BOOTPP_YIADDR_OFFSET+4)
#define BOOTPP_GIADDR_OFFSET (BOOTPP_SIADDR_OFFSET+4)
#define BOOTPP_CHADDR_OFFSET (BOOTPP_GIADDR_OFFSET+4)
#define BOOTPP_SNAME_OFFSET  (BOOTPP_CHADDR_OFFSET+16)
#define BOOTPP_FILE_OFFSET   (BOOTPP_SNAME_OFFSET+64)
#define BOOTPP_VEND_OFFSET   (BOOTPP_FILE_OFFSET+128)
#define BOOTPP_PACKET_SIZE   (BOOTPP_VEND_OFFSET+64)

/* OP codes */
#define BOOTP_OP_REQUEST 1
#define BOOTP_OP_REPLY 2

/* BOOTPP ethernet hardware type (same as for ARP except 1 byte) */
#define BOOTP_ETHERNET_HTYPE 1

/* UDP ports used by bootp */

static udp_port udp_port_bootp_server = {0, 67};
static udp_port udp_port_bootp_client = {0, 68};



/*
 * Static control data
 */

static int bootp_response_received;
static unsigned int xid;

/*
 * Internal functions
 */


/* bootp_input - handles received bootp packets
 *
 * Arguments:
 *     protocol_data - protocol data passed to module when protocol registered
 *     device_no     - device on which packet was received;
 *     packet        - packet, with UDP header removed;
 *     size          - size of packet;
 *     source_port   - source udp port of packet;
 *     source_address- source ip address of packet;
 *     frame_buffer  - original buffer containing packet;
 *     balance_buffer- buffer returned to maintain buffer balance.
 *
 * DESCRIPTION:
 *     Processes a bootp packet received on the client UDP port. Note that the client port is only registered
 *     as a UDP port when there is a bootp transaction in progress so I can assume the static data is valid.
 */
static void bootp_input(void *protocol_data, 
  int device_no, 
  unsigned char* packet, 
  int size, 
  udp_port source_port,
  udp_port dest_port, 
  ip_addr source_address, 
  unsigned char* frame_buffer, 
  unsigned char* *balance_buffer)
{
    int i;
    ip_addr my_original_ip_addr, my_ip_addr;
    mac_addr my_mac_addr;

    /* Set up the buffer to be returned (always frame buffer since this
     * function never keeps it or passes it on) */
    *balance_buffer = frame_buffer;

    /* Check that the packet is a reply packet; assuming everybody is obeying
     * the rules this is actually overkill since requests will be sent to the
     * server port. */
    if (*(unsigned char *)(packet + BOOTPP_OP_OFFSET) != BOOTP_OP_REPLY) {
		return;
    }

    /* Check that the xid matches */
    if ((*(unsigned char *)(packet + BOOTPP_XID_OFFSET) << 24 | 
      *(unsigned char *)(packet + BOOTPP_XID_OFFSET + 1) << 16 | 
      *(unsigned char *)(packet + BOOTPP_XID_OFFSET + 2) << 8 | 
      *(unsigned char *)(packet + BOOTPP_XID_OFFSET + 3)) != xid) {


	return;
    }

    /* For completeness check the hardware address type and size and that the
     * client IP and MAC addresses are correct */
    ip_get_device_address(device_no, my_original_ip_addr);
    ether_device_get_hw_address(device_no, my_mac_addr);

    if (*(unsigned char *)(packet + BOOTPP_HTYPE_OFFSET) != 
      BOOTP_ETHERNET_HTYPE || 
      *(unsigned char *)(packet + BOOTPP_HLEN_OFFSET) != MAC_ADDRESS_SIZE || 
      __rt_memcmp(my_original_ip_addr, packet + BOOTPP_CIADDR_OFFSET, 
      IP_ADDRESS_SIZE) != 0 || 
      __rt_memcmp(my_mac_addr, packet + BOOTPP_CHADDR_OFFSET, 
      MAC_ADDRESS_SIZE) != 0) {

	return;
    }

    /* This is the response we wanted; remember we have found it */
    bootp_response_received = TRUE;

	/* I'm not interested in the file, just the IP address of myself */

	for (i = 0; i < IP_ADDRESS_SIZE; i++)
	    my_ip_addr[i] = 
	      *(unsigned char *)(packet + BOOTPP_YIADDR_OFFSET + i);
	ip_set_device_addr(device_no, my_ip_addr);
}

/* bootp_boot - performs a bootp transaction
 *
 * Arguments:
 *     device_no       - device on which request should be sent
 *     input_file_name - file name supplied by caller; or null for default. Null terminated;
 *     server_addr     - returned server IP address
 *     boot_file_name  - returned boot file name supplied by server
 *     
 * Returned value:
 *     TRUE   - bootp successful; server_addr and boot_file_name valid;
 *     FALSE  - bootp failed; server_addr and boot_file_name invalid.
 */
int bootp_boot(int device_no)
{
    int i;
    ip_addr my_ip_addr;
    mac_addr my_mac_addr;
	unsigned char *current_bootp_buffer;
	int ticks;
    int send_count = 0;			/* Count of attempts at finding a bootp
					 * host */
    unsigned int retry_delay = 800;		/* Time, in zen, between attempts
					 */
    unsigned int time_passed = 0; 

	int err = 0;
    /* Find out who we are (if we know) */

    ip_get_device_address(device_no, my_ip_addr);
    ether_device_get_hw_address(device_no, my_mac_addr);

    /* Set up the xid; this should be the same for all attempts */
    xid =  0xf124b692; /* some random number */

    /* Remember that we haven't yet received a reply */
    bootp_response_received = FALSE;

    /* Tell udp that we want bootp client packets */
    udp_register_well_known_port(device_no, 
      udp_port_bootp_client, 
      &bootp_input, 
      NULL);

    /* Now make up to 4 attempts at sending a bootp packet */
	
	if (NULL == (current_bootp_buffer = framemalloc(ETHER_BUFFER_SIZE))) {
		return -1;
	}	
    
	while ((send_count++ < 5) && !bootp_response_received) {
		unsigned char* packet;

		packet = current_bootp_buffer + udp_data_offset();

		/* Set up a packet */

		for (i = 0; i < BOOTPP_PACKET_SIZE + udp_data_offset(); i++)
		    *(unsigned char *)(current_bootp_buffer + i)= 0;

		*(unsigned char *)(packet + BOOTPP_OP_OFFSET)= BOOTP_OP_REQUEST;
		*(unsigned char *)(packet + BOOTPP_HTYPE_OFFSET)= 
		  BOOTP_ETHERNET_HTYPE;
		*(unsigned char *)(packet + BOOTPP_HLEN_OFFSET)= MAC_ADDRESS_SIZE;
		*(unsigned char *)(packet + BOOTPP_HOPS_OFFSET)= 0;
		*(unsigned char *)(packet + BOOTPP_XID_OFFSET)= 
		  (unsigned char) ((xid & 0xff000000) >> 24);
		*(unsigned char *)(packet + BOOTPP_XID_OFFSET + 1)= 
		  (unsigned char) ((xid & 0x00ff0000) >> 16);
		*(unsigned char *)(packet + BOOTPP_XID_OFFSET + 2)= 
		  (unsigned char) ((xid & 0x0000ff00) >> 8);
		*(unsigned char *)(packet + BOOTPP_XID_OFFSET + 3)= 
		  (unsigned char) (xid & 0x000000ff);

		*(unsigned char *)(packet + BOOTPP_SECS_OFFSET)= 
		  (time_passed & 0xff00) >> 8;
		*(unsigned char *)(packet + BOOTPP_SECS_OFFSET + 1)= 
		  time_passed & 0x00ff;
		for (i = 0; i < IP_ADDRESS_SIZE; i++)
		    *(unsigned char *)(packet + BOOTPP_CIADDR_OFFSET + i)= 
		      my_ip_addr[i];
		for (i = 0; i < MAC_ADDRESS_SIZE; i++) {
		    *(unsigned char *)(packet + BOOTPP_CHADDR_OFFSET + i)= 
		      my_mac_addr[i];
		}

	   	*(unsigned char *)(packet + BOOTPP_FILE_OFFSET)= 0;

		/* All other fields can be left as zero */

		/* send it */

		
		err = udp_write(device_no, 
		  packet, 
		  udp_port_bootp_server, 
		  local_ip_broadcast_address, 
		  udp_port_bootp_client, 
		  BOOTPP_PACKET_SIZE, 
		  current_bootp_buffer, 
		  &current_bootp_buffer); 

		if (0 == err)  { // Packet sent okay
			/* Try to get a response */
			start_timer(3680);
			ticks = 0;
			while (!bootp_response_received && (ticks < retry_delay)) {			
				if (timer_expired()) {
					ticks++;
					start_timer(3680);
				}	
			    /* poll the network */
			    ethernet_process_one_packet(device_no, current_bootp_buffer, 
			      &current_bootp_buffer);
			}

			/* Set things up for next time round in case I need to go round again
			 */
			time_passed += (retry_delay >> 7); // or 10x there abouts
		}
	}

    /* Either I have received a response or I have run out of retrys; ether way
     * tell UDP we aren't interested in any more responses */
    udp_delete_port(device_no, udp_port_bootp_client);

    framefree(current_bootp_buffer);
	
	if (bootp_response_received) {

	/* The data from the response will have already been put in the right
	 * places; nothing more to do */
	return 0;
    }
      else {

	/* and exit */
	return -1;
    }
}

/* bootp_init_module - initialise the module 
 */
void bootp_init_module(void) 
{
return;
}
 


#endif
#endif
