/*****************************************************************************

Copyright 1993, 1994, 1995 Digital Equipment Corporation, Maynard, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted, provided
that the copyright notice and this permission notice appear in all copies
of software and supporting documentation, and that the name of Digital not
be used in advertising or publicity pertaining to distribution of the software
without specific, written prior permission. Digital grants this permission
provided that you prominently mark, as not part of the original, any
modifications made to this software or documentation.

Digital Equipment Corporation disclaims all warranties and/or guarantees
with regard to this software, including all implied warranties of fitness for
a particular purpose and merchantability, and makes no representations
regarding the use of, or the results of the use of, the software and
documentation in terms of correctness, accuracy, reliability, currentness or
otherwise; and you rely on the software, documentation and results solely at
your own risk.								   

******************************************************************************/

/*
 * MODULE DESCRIPTION:
 *
 *     DECCHIP_21040 Ethernet device  *
 * HISTORY:
 *
 * $Log: dec21040.c,v $
 * Revision 1.16  1997/11/13  15:29:57  smithg
 * 21143 csr fiddle
 *
 * Revision 1.15  1997/11/12 10:05:36  smithg
 * Added prototype
 *
 * Revision 1.14  1997/10/21 14:49:00  smithg
 * delay added
 *
 * Revision 1.13  1997/10/21 13:42:50  smithg
 * handles standalone, failed ethernet card and functional
 * ethernet card cases for 21040, 21140
 *
 * Revision 1.12  1997/10/20 13:02:17  smithg
 * 21140 okay
 *
 * Revision 1.11  1997/10/17 10:26:33  smithg
 * Better error handling
 * lies to Angel about device status
 * cos Angel ignores errors from initiatisation
 *
 * Revision 1.10  1997/10/15 14:26:46  smithg
 * 21040 -> 21x4x
 *
 * Revision 1.9  1997/10/14 17:51:21  smithg
 * no debug, timer.h included
 *
 * Revision 1.7  1997/10/13 09:12:25  smithg
 * 21143 not functioning
 *
 * Revision 1.6  1997/10/02 14:19:10  smithg
 * changed negotiation for 21143 - won't work
 * 21040 works
 *
 * Revision 1.5  1997/09/15 12:39:32  smithg
 * more tweaks - functions
 *
 * Revision 1.4  1997/08/21 14:09:07  smithg
 * 21285 working 1
 *
 * Revision 1.3  1997/08/15 12:18:31  smithg
 * Force 10BaseT on 21040
 *
 * Revision 1.2  1997/07/23 18:06:17  smithg
 * checkpoint
 *
 * Revision 1.1  1997/07/23 16:16:11  smithg
 * Initial revision
 *
 * Revision 2.0  1997/07/23 13:26:04  smithg
 * ebsa285 revisions
 *
 * Revision 1.2  1997/07/23 10:57:10  smithg
 * Header updated
 *
 * Revision 1.1  1997/07/18 12:21:07  smithg
 * Initial revision for Angel
 */

#include "ether.h"
#include "devconf.h"


#if (ETHERNET_SUPPORTED == 1)
#include "dec21040.h"
#include "system.h"
#include "pci.h"
#include "edriver.h"
#include "net_buff.h"
#include "timer.h"
#include "errors.h"
#include "support.h"

extern U32 SAr_PciBenchMode(void);

/* MODULE DESCRIPTION:
 *
 * This file contains DECCHIP_21040 specific routines needed for an Ethernet
 * driver. The driver follows the generic EB64 Ethernet architecture functions
 * with support for the following routines:
 *
 *	    o ether_device_init()
 *	    o ether_device_read()
 *	    o ether_device_write()
 *	    o ether_device_flush()
 *	    o ether_device_print_stats()
 *	    o ether_device_clear_interrupts()
 *	    o ether_device_get_hw_addr()
 *
 *	    o ether_device_register()
 *	    
 * Please see the individual routines for the parameter listings and any
 * returned values.
 *
 * When a device is installed (registered) a structure including pointers to
 * these functions will be built. The structure is part of an array, the array
 * entry being dictated by the device number. While these routines have 
 * prefixes of AMD79C960, they will be called from the registered array using
 * the generic "ether" prefixed routines.
 * 
 * Supporting routines include:
 *
 *        Routines to read and write the RDP and IDP registers for CSR IO
 *        (these include implicit RAP register control)
 *
 *        Start the device and reset the counter array
 *
 *        A selection of utilities to print packets (RX and TX), print
 *        the counter array, and dump the CSR's.
 *
 * The main functions are:
 *
 *        Ethernet initialisation - set up initialisation block, set up
 *        RX and TX descriptor rings, allocate buffer space and define 
 *        arrays of pointers to the TX and RX buffers, then initialise
 *        the device.
 *
 *        Receive a packet in a single buffer
 *
 *        Send a packet from a single buffer
 *
 *	  Take control (with a watchdog) until all queued packets sent
 *
 *        Clear down the daisy chained interrupt in the 82C486 ISA controller
 *
 *	  Fetch the associated MAC address
 *
 *	  Register a new device in the Ethernet dispatch table
 *
 * Assumptions are:
 *
 *        All buffers can hold maximum length packets - no chaining
 *
 *        NOTE: static allocation of buffer memory means that the
 *              maximum number of DECCHIP_21040 must be declared at compile time
 *	        using DECCHIP_21040_MAX_DEVICES. IO base addresses are also fixed
 *              for all of these at compile time. Device registration has to 
 *              ensure there is no overlapping IO address ports between 
 *	        different device types if the possibility exists at compile 
 *		time.
 *
 *	  Once a driver is running and sending/receiving traffic, pointers
 *	  to buffers are then swapped with the datalink layer, meaning that
 *        the number of buffers allocated to the routine remains constant,
 *	  but the actual buffers statically assigned at compile time migrate
 *	  through the system.
 *
 *	  An array is used to map device numbers to ISA bus base addresses.
 *	  Entries are added to the array when a device is registered and
 *	  allocated a device number. 
 *
 *        PLEASE NOTE: This array is currently Ethernet specific.
 *
 *        Global parameters are used to set up the following:
 *
 *                etherdump - dictates whether packets sent and received
 *                            are output to the monitor using printf()
 *
 */

/* DECCHIP_21040_MAX_DEVICES copies of resources are statically assigned at compile time.  */

/*
 * Control block and data memory requirements are as follows:
 *
 *        Initialisation block
 *        RX descriptors (number defined as RRING_SIZE)
 *        TX descriptors (number defined as SRING_SIZE)
 *        RX buffers (number defined as RRING_SIZE; size of BUFFER_SIZE)
 *        TX buffers (number defined as SRING_SIZE; size of BUFFER_SIZE)
 *
 * Two pointer arrays and index variables are created for buffer manipulation:
 *
 *        current_rring_buffers[][] for RX
 *        current_sring_buffers[][] for TX (send)
 */
 

/*
	Map the SAaddress to PCI space
	As far as the tulip is concerned the address to write to will be
	(memBAR of 21285) + (offset from PCI base in SA memory)
	Say 0x01000000 was PCI base in SA memory
	And the area assigned to Tulip starts at base + 0xa000
	
	The Tulip will need to write to (memBAR of 21285) + 0xa000
	Don't know memBAR 0f 21285
*/
 
/*
 * A description of all the DECchip 21040 devices in this system.
 * Information is filled in via DECchip_21040_device_init_module().
 */
static DECchip_21x4x_device_t devices[DECCHIP21040_MAX_DEVICES];
/* 
 * Allocated by DECchip_21x4x_device_init_module().  Buffers are
 * only allocated for valid devices.
 */

static unsigned char *current_rring_buffers[DECCHIP21040_MAX_DEVICES][RRING_SIZE];
static unsigned char *current_sring_buffers[DECCHIP21040_MAX_DEVICES][SRING_SIZE];

/* 
 * Statically allocate all of the rx and tx descriptors (there aren't
 * too many).
 */

static DECCHIP_21x4x_RX_BD *rring[DECCHIP21040_MAX_DEVICES];
					/* pointer to RX ring descriptor memory
					 * for all potential devices */
static DECCHIP_21x4x_TX_BD *sring[DECCHIP21040_MAX_DEVICES];
					/* pointer to TX ring descriptor memory
					 * for all potential devices */

/*
 	SROM Structure
	Here we only care about the first controller,
	hence the srom struct format
*/
static struct de4x5_srom {
    char id_block[18];
    char version;
    char num_adapters;
    char hw_addr[6];
	char device_no; // ignore on single ctrl devices  
	char leaf_offset[2]; // offset from start of Srom
    char info[97];
    short chksum;
  } Srom;

/* 
 * Functions declared within 
 */
static DECchip_21x4x_device_t *DECchip_21x4x_device_find(int device_no);
static int DECchip_21143_Reset_and_Setup(DECchip_21x4x_device_t *device);
static int DECchip_21x4x_device_init(int device_no);
static int DECchip_21x4x_device_read(int device_no, 
  unsigned char balance_buffer[ETHER_BUFFER_SIZE], 
  unsigned char **frame_buffer);

static int DECchip_21x4x_device_write(int device_no, unsigned char *frame, 
  int size, unsigned char frame_buffer[ETHER_BUFFER_SIZE], 
  unsigned char **balance_buffer);
static int DECchip_21x4x_device_flush(int device_no);
static void DECchip_21x4x_device_print_stats(int device_no);
static void DECchip_21x4x_device_clear_interrupts(int device_no);
static int  DECchip_21040_device_get_hw_address(int device_no, 
  mac_addr hw_address);
#if 0  
static void DECchip_21x4x_device_stop(int device_no);
#endif
static void DECchip_21x4x_device_preg(int device_no);
					/* dump all registers */

static int DECchip_21x4x_device_check_romid(unsigned char * romid);
static void DECchip_21x4x_device_allocate_buffers(DECchip_21x4x_device_t * device);
static int DECchip_21x4x_device_hw_address_valid(mac_addr hw_address);

static int DECchip_21x4x_device_get_hw_address(int device_no, 
  mac_addr hw_address);
static void read_entire_srom(DECchip_21x4x_device_t *device);

// Global to indicate success state of ethernet stack 
extern int InitialisationFail;

static struct {
  int DeviceID;
  char *DeviceIDString;
} DC21X4X_DID_Data[] = { 
  { DECCHIP_21040_DID, "DC21040" },
  { DECCHIP_21041_DID, "DC21041" },
  { DECCHIP_21140_DID, "DC21140" },
  { DECCHIP_21142_DID, "DC21142" },
  { DECCHIP_21143_DID, "DC21143" },
  { 0, NULL }
};


#if 0
/*
	Debug routine
	CSRs are 32 bits wide but 64 bit aligned.
*/
void print_CSRs(void)
{
 DECchip_21x4x_device_t *device = DECchip_21x4x_device_find(0);
 PU32 ptr;	
 int i;	
 
 	putstring("\nCSRs:\n");
	ptr = (PU32)device->PCI_IO_Base;
	put(device->PCI_IO_Base);
	putstring("\n");
	for (i=0;i<16;i++) {
		if((i % 4) == 0)
			putstring("\n");
		put(ptr[i*2]);
		
	}
}


void print_desc()
{
 DECchip_21x4x_device_t *device = DECchip_21x4x_device_find(0);
 U32 base, *ptr;	
 int i;	
U32 bar, pciBase, pciSize;
devNodeType	*node ;

 		node = SAr_FindHostNode() ;

		if (node != NULL) {
			pciSize = node->memSize[HOST_MEMBAR] ;
			bar = HOST_MEMBAR ;
			pciBase = SAr_GetNextMemBAR(node->handle, &bar) ;
		}
 	putstring("\nTX desc:\n@");
	base = *(U32*)(device->PCI_IO_Base + CSR4);
	ptr = (U32*)( base - pciBase + (U32)PCI_DRAMBASE );
	put((int)ptr);
	for (i=0;i<4;i++) {
		put(ptr[i]);		
	}
 	putstring("\nRX desc:\n@");
	base = *(U32*)(device->PCI_IO_Base + CSR3);
	ptr = (U32*)( base - pciBase + (U32)PCI_DRAMBASE );
	put((int)ptr);
	for (i=0;i<4;i++) {
		put(ptr[i]);		
	}		

}
#endif

/* DECchip_21x4x_device_find()
 *
 * Arguments:
 * 
 *    device_no    - device number 
 *
 * Returned Value:
 *    address of device or NULL if failed.
 */
static DECchip_21x4x_device_t *DECchip_21x4x_device_find(int device_no)
{
    register int i;
    register DECchip_21x4x_device_t *device = devices;

    for (i = 0; i < DECCHIP21040_MAX_DEVICES; i++) {
	if ((int)(device->exterior_device) == device_no)
	    return (device);
	else
	    device++;
    }
    return (NULL);
}



static void DECchip_21x4x_Reset( DECchip_21x4x_device_t *device)
{
S32 status;
U32 index = device->index;

	/* 
		A hidden bit of the data sheet suggests the card is still asleep,
		Wake up you lazy twat!!!!!!!!!!!!!!!
		Config reg CFDD @ 0x40, bit 31 is sleep, bit 30 is snooze.
				
	*/

	// Bus mode register - sw reset
 	_21x4xWriteCSR(device, CSR0, CSR0_SWR);
	_msleep(1);


	// Wake up sleepy head
	SAr_PciCfgWriteLong(device->pci, (U32)0x40, (U32)0, (PS32)&status);
	_sleep(1);
	

	// Bus mode register - cache alignment set to 1,1 - 32-word
 	_21x4xWriteCSR(device, CSR0, 0x00c0c000);
	_msleep(1);
		
	// TX ring address       
	_21x4xWriteCSR(device, CSR4,  (unsigned int)(SAr_Mem2Pci((void *)sring[index], (U32*)&status)));

	// RX ring address
    _21x4xWriteCSR(device, CSR3, (unsigned int)(SAr_Mem2Pci((void *)rring[index], (U32*)&status)));

	
	// Interrupts off
	_21x4xWriteCSR(device, CSR7, 0);
	// Stop SIA
    _21x4xWriteCSR(device, CSR13, 0x00000000);
					

	     

}



static int DECchip_21x4x_Setup_Rings(DECchip_21x4x_device_t *device)
{
int i;
U32 status;
U32 index = device->index;
static setup_done = FALSE;

	if (!setup_done) {	
	 	// Set up the transmit and receive rings  for current device_index.
	    for (i = 0; i < RRING_SIZE; i++) {	/* For all RX descriptor entries */
			__rt_memset(&rring[index][i], 0, sizeof(DECCHIP_21x4x_RX_BD));	/* clear descriptor */
			rring[index][i].u_rdes0.s_rdes0.OWN = TRUE;	/* 21040 owns it */
		
		 	// If this is the end of the ring, mark it as so.	 														 
			if (i == (RRING_SIZE - 1))
					    rring[index][i].u_rdes1.s_rdes1.RER = TRUE;
		
		 	// Point at the receive buffer and give its length.	 
			rring[index][i].u_rdes1.s_rdes1.RBS1 = ETHER_BUFFER_SIZE; // RBS2 not set - hence ignore buffer 2
			status = 0;
			rring[index][i].rdes2 = (int)SAr_Mem2Pci((void*)current_rring_buffers[index][i],(U32*) &status);
			if (OK != status) {
				return ERROR;
			}
	    }

		
	    for (i = 0; i < SRING_SIZE; i++) {	/* For all TX descriptor entries */
			__rt_memset(&(sring[index][i]), 0, sizeof(DECCHIP_21x4x_TX_BD));
							/* clear descriptor */
			sring[index][i].u_tdes0.s_tdes0.OWN = FALSE;
							/* We own it */
			sring[index][i].u_tdes1.s_tdes1.LS = TRUE;	// This is the last segment in the frame
			sring[index][i].u_tdes1.s_tdes1.FS = TRUE;	// Its the first too
		
		 	// If this is the last descriptor in the ring, then mark it as so.
			if (i == (SRING_SIZE - 1))
			    sring[index][i].u_tdes1.s_tdes1.TER = TRUE;
		/*
		 * tdes2 contains a pointer to the transmit buffer so point it
		 * and set up its size in tdes1.
		 */
			sring[index][i].u_tdes1.s_tdes1.TBS1 = ETHER_BUFFER_SIZE;
			sring[index][i].tdes2 = (unsigned int)SAr_Mem2Pci((void*)current_sring_buffers[index][i],(U32*) &status);
			if (OK != status) {
				return ERROR;
			}
	    }
	setup_done = TRUE;	
	} 
    	device->rx_index = 0;		/* Reset RX array index */
	    device->tx_index = 0;		/* Reset TX array index */
		// TX ring address       
		_21x4xWriteCSR(device, CSR4,  (unsigned int)(SAr_Mem2Pci((void *)sring[index], (U32*)&status)));

		// RX ring address
	    _21x4xWriteCSR(device, CSR3, (unsigned int)(SAr_Mem2Pci((void *)rring[index], (U32*)&status)));
	
	return 0;
}



/*
 *	int select_media 
 *	The media available on each card will vary, so check all types that the chip 
 *	can handle.
 *		
 *	returns 
 *		- 0 if valid media and link found
 *		- non-zero on error
 */

int select_media(DECchip_21x4x_device_t *device)
{  // 21x4x link test
int LinkPass=ERROR;
U32 csr12;
		switch(device->device_type) {
				
		case DECCHIP_21040_TYPE:
		case DECCHIP_21041_TYPE:
			_21x4xWriteCSR(device, CSR7, 0x00000000);
					/* Int mask = no interrupt */
			_21x4xWriteCSR(device, CSR13, 0x00000000);
					/* Stop SIA */
			/*
			 *  First try a TP (10 Base T) connection (ie BIT 3, AUI = 0).
			 *  Then wait a little while and read the CSR12 LinkFail status
			 *  bit.  If this is set, then the interface has failed, so
			 *  try AUI.
			 */
			
			_21x4xWriteCSR(device, CSR13, (CSR13_CAC | CSR13_SRL));
			_sleep(2);
			
			csr12 = _21x4xReadCSR(device, CSR12);
			if ((csr12 & CSR12_LKF) != 0) {
			    /* Stop SIA */
			    _21x4xWriteCSR(device, CSR13, 0x00000000);
			 	_21x4xWriteCSR(device, CSR13, (CSR13_AUI | CSR13_CAC | CSR13_SRL));
			} else {
				LinkPass = 0;
			}
			
				
		break;
		
		case DECCHIP_21140_TYPE:
			 // Use SIA on 10BaseT
			 _21x4xClearCSRBits(device, CSR6, CSR6_PS);
			 _usleep(50);
			 // This device needs a reset to get it up again
			 DECchip_21x4x_Reset(device);
			 _21x4xClearCSRBits(device, CSR6, CSR6_PS);			 
		break;

	case DECCHIP_21143_TYPE:
	case DECCHIP_21142_TYPE:
	  /* THis is now taken care of in the specific setup function */
	    break;
	}// end of switch
		
	 return LinkPass;
}


static int DECchip_21143_Reset_and_Setup(DECchip_21x4x_device_t *device)
{
int LinkPass=ERROR;
U32 csr12;
S32 status;
U32 index = device->index;


/* the Pseudo code is from Alex - I'll try and follow it totally */


/* ! Enable master operation in the PCI slot
   write conf cfcs 47
   ! Set the Latency Timer
   write conf cflt ff00
   read conf cfid
   !
   */

/* The above is carried out in pci.c except that cfcs is set to Master plus 
   some memory setup */

/*! Reset the adapter
  write reg 0 1
  delay   500
  write reg 0 9000
  delay   500
  read reg 0
  !
  */


/*! Run in normal mode
  write config cfda 0
  !
  ! Start the receive and transmit lists
  List Address */


/* Bus mode register - sw reset */
 _21x4xWriteCSR(device, CSR0, CSR0_SWR);
 _msleep(500);

/* Bus mode register - sw reset */
 _21x4xWriteCSR(device, CSR0, 0);
 _msleep(500);

  /* Bus mode register - cache alignment set to 1,1 - 32-word */
 _21x4xWriteCSR(device, CSR0, 0x0000c000);
 _msleep(500);

 /* Wake up sleepy head */
 SAr_PciCfgWriteLong(device->pci, (U32)0x40, (U32)0, (PS32)&status);
 _sleep(1);
	
 /*! Start the receive and transmit lists
   /* TX ring address */  
 _21x4xWriteCSR(device, CSR4,  (unsigned int)(SAr_Mem2Pci((void *)sring[index], (U32*)&status)));

 /* RX ring address */
 _21x4xWriteCSR(device, CSR3, (unsigned int)(SAr_Mem2Pci((void *)rring[index], (U32*)&status)));

 if (ERROR == DECchip_21x4x_Setup_Rings(device))
     return ERROR;


/*!
  ! Set operation mode: Tx threshold = 160, operation mode = normal
  !                     no full duplex, pass all mcast, promiscuous mode,
  !                     pass bad frames, port select = 1 - SIO in hold
  write register 6 43C0c0c8 */


  _21x4xWriteCSR(device, CSR6, 0x43c0c0c8);	
 _usleep(2);

 /*  !
     ! disable event interrupts
     write register 7 0
     !  */

/* Interrupts off */
 _21x4xWriteCSR(device, CSR7, 0);
  _usleep(2);


  /*! Release watchdog timer and clear GPR.
    ! Set up GPR as output port, pins GEP0,GEP2 active low PHY reset, Power down */
  /*write register 15 08Af0000 */

  _21x4xWriteCSR(device, CSR15, 0x08af0000);
  _usleep(2);

  /*write register 15 00AF0000 */

  _21x4xWriteCSR(device, CSR15, 0x00af0000);
  _usleep(2);

  /* ! Reset PHY
     write register 15 00040000 */

  _21x4xWriteCSR(device, CSR15, 0x00040000);
  /* delay     400ms */
  _msleep(400);

  /*! Start PHY
    write register 15 00050000
    !*/
  _21x4xWriteCSR(device, CSR15, 0x00050000);
  _usleep(2);

  /* ! Enable SIO for 10baseT operation
     write register 13   */

  _21x4xWriteCSR(device, CSR13, 0x0001);
  _usleep(2);


  /*write register 14 00007f3f */

  _21x4xWriteCSR(device, CSR14, 0x00007f3f);

  /*delay     400 */
  _msleep(400);

  /*read register 13
    !
    ! Start rx,tx - same configuration as before, only start rx and tx and phy
    write register 6 43C0e0ca
    ! */


  _21x4xWriteCSR(device, CSR6, 0x43c0e0ca);
  _msleep(400);


  /*!
    ! Twinnet2, QSI symPHY operation, 10BASET half duplex.
    !      */

  LinkPass = TRUE;

return LinkPass;

}



/* DECchip_21x4x_device_init - initialises an ethernet device
 *
 * Arguments:
 *
 *    device_no    - device number
 *
 * Returned value
 *    0         - initialised succesfully
 *    -ve        - error on initialisation
 */

static int DECchip_21x4x_device_init(int device_no)
{
	S32 status;
    DECchip_21x4x_device_t *device = DECchip_21x4x_device_find(device_no);
    U32 csr6;
    U32 index;
	int i,j, setup_attempts = 5;


    if (device == NULL) {
		return ERROR;
	}
  else if (DECCHIP_21143_TYPE == device->device_type)
    {
      /* Let's make the setup of the 21143 an individual routine */

      DECchip_21143_Reset_and_Setup(device);

      /* Still need to get the Hardware address - removed by the if */
      DECchip_21x4x_device_get_hw_address(device_no, device->hw_address);

      if (FALSE == DECchip_21x4x_device_hw_address_valid(device->hw_address)) 
       	  return ERROR;

     return 0;

    }
  else
    {
/*
 *  If this device doesn't exist or hasn't been allocated, then
 *  return an error.  Otherwise, get the device index.
 */

    if (!(device->flags.exists) || !(device->flags.allocated)) {
		return ERROR;
	}
    index = device->index;
	
	// Wake it up
	DECchip_21x4x_Reset(device); 

if (ERROR == DECchip_21x4x_Setup_Rings(device))
	return ERROR;

	/*
	 * Try and figure out this device's hardware (mac) address, if
	 * we cannot figure one out, then fail the initialisation.
	 */
 	// Is it a 21040 or otherwise?
    if (device->device_type == DECCHIP_21040_TYPE)
      	DECchip_21040_device_get_hw_address(device_no, device->hw_address);
    else {
		DECchip_21x4x_device_get_hw_address(device_no, device->hw_address);
	}

    if (FALSE == DECchip_21x4x_device_hw_address_valid(device->hw_address)) {
	
      	return ERROR;
	}

    


    if (device->flags.loopback) {
		/* Stop SIA */	
		_21x4xWriteCSR(device, CSR13, 0x00000000);
		/* EXT SIA                 */
		_21x4xWriteCSR(device, CSR15, 0x00000000);
		/* EXT SIA                 */
		_21x4xWriteCSR(device, CSR14, 0x00000000);
		/* EXT SIA (Was 0x00003041)*/
		_21x4xWriteCSR(device, CSR13, 0x00000001);
		// Interrupts off
		_21x4xWriteCSR(device, CSR7,  0x00000000);
    } else {
		/*	Use select media to find the connection, if it fails then initialise anyway,
			the card may recover when a lead is attached to it.
		*/
		select_media(device);
	}
    }

   	if (device->flags.loopback) {
		_21x4xWriteCSR(device, CSR6, 0x00002542);
	}				 
    else {
	if (device->flags.setup_frame) { // Build the setup frame
		/*
		 *  We're building a perfect filter setup with two entries
		 *  (the broadcast address and our hardware address).
		 */
		int k;


		// Empty the frame....
	    __rt_memset(device->setup_frame, 0x00, SETUP_FRAME_SIZE);
		k =0;
		while (k < 12)
			device->setup_frame[k++] = (char)0xff;
		//  Slap in our physical address in the rest 
		while (k < SETUP_FRAME_SIZE) {
			for (i=0;i<6;i+=2){
				for (j=0;j<2;j++) {
					device->setup_frame[k++] =
					 device->hw_address[i+j];
				}
			k+=2;	
			}
		}

		/*
		 *  Now send it to the 21040.  Note that this transmit buffer
		 *  will *not* get recycled !!correct!! as its address never gets put into
		 *  current_sring_buffers.  However, we do keep the address of
		 *  it in the device data structure so that we can use it when
		 *  debugging.  Also note that we do not increment the tx index,
		 *  this is because we wait for the frame to be returned to
		 *  us before leaving this procedure.
		 */

	    sring[device->index][device->tx_index].u_tdes1.s_tdes1.SET = TRUE;
	    sring[device->index][device->tx_index].u_tdes1.s_tdes1.TBS1 = SETUP_FRAME_SIZE;
	    sring[device->index][device->tx_index].u_tdes1.s_tdes1.LS = 0;
	    sring[device->index][device->tx_index].u_tdes1.s_tdes1.FS = 0;
		
		
		//  Perfect address filtering on
	    sring[device->index][device->tx_index].u_tdes1.s_tdes1.HP = FALSE;
	    sring[device->index][device->tx_index].u_tdes1.s_tdes1.IV = FALSE;
		
	    // Set setup frame address
		sring[device->index][device->tx_index].tdes2 = (unsigned int) (
	      SAr_Mem2Pci((void *) device->setup_frame, (U32*)&status));
	}


	// Okay, lets get this show on the road
	csr6 = _21x4xReadCSR(device, CSR6);
	
	// Make sure TX/RX are stopped 
	if (csr6 & (CSR6_ST | CSR6_SR)) {
		csr6 &= ~(CSR6_ST | CSR6_SR);
		_21x4xWriteCSR(device, CSR6, csr6);
	}

	// don't pass bad frames 
	csr6 &= ~(CSR6_PB);

	//	Give packet to 21x4x
    sring[device->index][device->tx_index].u_tdes0.s_tdes0.OWN = TRUE;

	if (device->flags.setup_frame) {	
		// Setup frame sent, wait before starting recv.
		
    	csr6 &= ~CSR6_PR;	/* don't run in promiscuous mode. */

		/*
			By now CSR4 points to TDesc0 of the 0th descriptor
			Tdesc2 points to the setup frame
			tdes1 has size 192
			tdes0 has ownership bit set ie 21x4x owns it.
			So start TX		
		*/
 
    	_21x4xWriteCSR(device, CSR6, csr6 | CSR6_ST);
	 	// Wait until the setup frame comes back.
		while ((setup_attempts-- > 0) && 
			(sring[device->index][0].u_tdes0.s_tdes0.OWN == TRUE)) {
			_msleep(2);
		}
		
		if (!setup_attempts) {
			return ERROR;
		}
		device->flags.initialised = TRUE;
		
	 	// Fixup the descriptor.
			 
	    sring[device->index][device->tx_index].u_tdes0.tdes0 = 0;
	    sring[device->index][device->tx_index].u_tdes1.tdes1 = 0;
	    sring[device->index][device->tx_index].u_tdes1.s_tdes1.TBS1 = ETHER_BUFFER_SIZE;
	    sring[device->index][device->tx_index].u_tdes1.s_tdes1.LS = TRUE;
	    sring[device->index][device->tx_index].u_tdes1.s_tdes1.FS = TRUE;
	    sring[device->index][device->tx_index].tdes2 = (unsigned int) 
	      SAr_Mem2Pci((void *) current_sring_buffers[device->index][device->tx_index], (U32*)&status);
		
	/*
	 *  Increment the transmit index as the Tulip now
	 *  believes that it is on transmit descriptor 1 (*not* 0).
	 */		    
	    device->tx_index++;
	
	/*
	 *  Now start receive.
	 */
	    _21x4xWriteCSR(device, CSR6, csr6 | CSR6_ST | CSR6_SR);
	} else
	    _21x4xWriteCSR(device, CSR6, csr6 | CSR6_ST | CSR6_PR | CSR6_SR);
    }
    return 0;
}


static int check_rx_desc(DECchip_21x4x_device_t *device)
{
DECCHIP_21x4x_RX_BD *rxdesc;

	rxdesc = (DECCHIP_21x4x_RX_BD*)&rring[device->index][device->rx_index];
	if (rxdesc->u_rdes0.s_rdes0.ES) { // there's an error
#if 0
		putstring("\nRX error:");
		if (rxdesc->u_rdes0.s_rdes0.OF)
			putstring("\nOverflow");
		if (rxdesc->u_rdes0.s_rdes0.CE)
			putstring("\nCRC");
		if (rxdesc->u_rdes0.s_rdes0.CS)
			putstring("\nCollision seen");
		if (rxdesc->u_rdes0.s_rdes0.TL)
			putstring("\nFrame too long");
		if (rxdesc->u_rdes0.s_rdes0.RF)
			putstring("\nRunt frame");
		if (rxdesc->u_rdes0.s_rdes0.LE)
			putstring("\nLength error");
#endif
	return ERROR;	
	}
return 0;
}



/* DECchip_21x4x_device_read - reads a frame from an Ethernet device
 *
 * Arguments:
 *
 *    device_no      - device number
 *    balance_buffer - spare buffer to maintain buffer balance
 *    frame_buffer   - returned buffer containing frame
 *
 * Returned value:
 *
 *    positive     - size of frame read; frame copied into buffer if <= size.
 *       0         - nothing to read
 *    negative     - error on read
 *
 * DESCRIPTION:
 *    This function reads a frame from the device if there is one ready to be 
 *    read.  The frame read is returned in frame_buffer.  To maintain the 
 *    driver's pool of buffers
 *    the caller must give the driver an unused buffer (balance_buffer) before
 *    the
 *    driver returns the frame.  Once this function has been called 
 *    balance_buffer must
 *    not be used by the calling software.
 *
 *    On return frame_buffer will always point to a valid buffer.  If the read 
 *    has failed
 *    (result <= 0) the contents of this buffer will not be valid.  Note that
 *    whether or not the read succeeds it is undefined whether the same buffer
 *    will be returned as
 *    frame_buffer as was passed down as balance_buffer.
 */

static int DECchip_21x4x_device_read(int device_no, 
  unsigned char balance_buffer[ETHER_BUFFER_SIZE], 
  unsigned char **frame_buffer)
{

    DECchip_21x4x_device_t *device = DECchip_21x4x_device_find(device_no);
    int count;
    U32 csr5;
	U32 status;
/* 
 *  Set up initial return value for error cases 
 */


	
    *frame_buffer = balance_buffer;
/*
 *  Valid device?
 */
    if (device == NULL) {
		return ERR_INVALIDDEVICENUM;
    }
/*
 *  If the receive process has stopped or is suspended, then restart it.
 */
    csr5 = _21x4xReadCSR(device, CSR5);
    if (((csr5 & CSR5_RPS) != 0) || ((csr5 & CSR5_RS) == CSR5_RS_SUSPENDED)) {
		_21x4xWriteCSR(device, CSR2, CSR2_RPD);
    }
/*
 *  Keep going until we either get a good receive or the receive
 *  descriptor is owned by the 21040.
 */

    while (rring[device->index][device->rx_index].u_rdes0.s_rdes0.OWN == 
      FALSE) {

	/*
 	 *  We've got a received buffer, check for errors reported by the
 	 *  device.  First check for the Error Summary bit if it is FALSE,
 	 *  all is good, else check further into the error.
 	 */
	if (check_rx_desc(device)) {
		/*
		 *  Clear the whole of the descriptor.
		 */
	    rring[device->index][device->rx_index].u_rdes0.rdes0 = 0;
		/*
		 *  Set the count to -1 to indicate an error.
		 */
	    count = ERR_CHKRECVERR;
	} else { // Okay, buffer received.
	    count = 
	      rring[device->index][device->rx_index].u_rdes0.s_rdes0.FL - 4;
					/* -4 deducts the CRC bytes */

		/* 
		 *  Check for errors.  For all errors, set count negative. If no errors,
		 *  copy the frame to the caller buffer. 
		 */
	    if (count > ETHER_BUFFER_SIZE) {
			count = ERR_ETHERPACKETTOOLONG;
		}
	    else {
			*frame_buffer = 
		  	 current_rring_buffers[device->index][device->rx_index];
#if CACHE_SUPPORTED
  			Cache_IBR((char*)frame_buffer, ((char*)frame_buffer)+ETHER_BUFFER_SIZE);
#endif
			 
					/* return pointer to buffer with the
					 * frame in it */
			current_rring_buffers[device->index][device->rx_index] = balance_buffer;
					/* swap in free buffer supplied */
			rring[device->index][device->rx_index].rdes2 = (int)
			  	 SAr_Mem2Pci((void*) balance_buffer, (PU32)&status);
			if (status != OK) {
				count = ERR_MEM2PCIFAIL;	
			} 
	    }
	}				
	/* 
	 *  Give the buffer back to the 21040. Advance the ring index to the next
	 *  buffer. 
	 */
	rring[device->index][device->rx_index].u_rdes0.s_rdes0.OWN = TRUE;
		device->rx_index = (device->rx_index + 1) % RRING_SIZE;
	/*
	 *  If we have a valid count, then return to the caller.
	 */

	return count;
	
    }					/* end of while descriptors are ours */
	/*
	 *  We didn't find anything to read (but we may have drained part if
	 *  not all of the receive ring).  Return 0 to the caller.
	 */
    return 0;
}

static int check_tx_desc(DECchip_21x4x_device_t *device)
{
DECCHIP_21x4x_TX_BD *txdesc;

	txdesc = (DECCHIP_21x4x_TX_BD*)&sring[device->index][device->tx_index];
	if (txdesc->u_tdes0.s_tdes0.ES) { // there's an error
#if 0
		putstring("\nTX error:");
		if (txdesc->u_tdes0.s_tdes0.UF)
			putstring("\nUnderflow");
		if (txdesc->u_tdes0.s_tdes0.LF)
			putstring("\nLinkFail");
		if (txdesc->u_tdes0.s_tdes0.EC)
			putstring("\nExcessive Collisions");
		if (txdesc->u_tdes0.s_tdes0.LC)
			putstring("\nLate collision");
		if (txdesc->u_tdes0.s_tdes0.NC)
			putstring("\nNo carrier");
		if (txdesc->u_tdes0.s_tdes0.LO)
			putstring("\nLoss of carrier");
			
		if (txdesc->u_tdes0.s_tdes0.TO)
			putstring("\nJabber");
#endif
	return ERROR;	
	}
return 0;
}



/* DECchip_21x4x_device_write - queue a frame to be sent on an ethernet device
 * Arguments:
 *     device_no         - device number    
 *     frame             - pointer to start of frame within the buffer    
 *     size              - size of frame    
 *     frame_buffer      - buffer containing frame    
 *     balance_buffer    - returned buffer to maintain buffer balance
 * Returned value:
 *    TRUE - successfully queued for sending    FALSE - unable to send
 * Description:
 *
 * This function writes a frame to the device  To maintain the caller's pool of
 * buffers    the driver must give the caller an unused buffer (balance_buffer)
 * in return for the buffer containing the frame to be transmitted.
 * If the send succeeds then frame_buffer must not be accessed by the caller
 * after this  call.  If the send can't be queued then frame_buffer will remain
 * valid and the returned value of balance_buffer is undefined and must not 
 * be used.
 */
static int DECchip_21x4x_device_write(int device_no, unsigned char *frame, 
  int size, unsigned char frame_buffer[ETHER_BUFFER_SIZE], 
  unsigned char **balance_buffer)
{
    DECchip_21x4x_device_t *device = DECchip_21x4x_device_find(device_no);
    U32 csr5;
	U32 status;
/*
 * Valid device?
 */

    if (device == NULL) {
		return ERR_INVALIDDEVICENUM;
    }


/* 
 * Exit if next buffer is owned by the 21040. Enforce minimum packet
 * length. Check the buffer for previous errors. Swap buffer into the lance
 * ring and associated buffer pool.  
 */
    if (sring[device->index][device->tx_index].u_tdes0.s_tdes0.OWN)
		return ERR_TXRINGFULL;			/* TX ring full - cannot send */

    if (size < MIN_ENET_PACKET)
		size = MIN_ENET_PACKET;		// ensure minimum pkt size criteria met
					 
    if (size > MAX_ENET_PACKET) {	// flag an error if the pkt is too big
					 
		return ERR_ETHERPACKETTOOLONG;
    }
/*
 *  Check that there wasn't an error the last time we sent using this 
 *  descriptor.   Deleted
 */
    
/*
 *  Put the transmit buffer passed into the current set of transmit
 *  ring buffers and return as the balance buffer the one that the
 *  transmit buffer replaces (ie swap them).
 */
    *balance_buffer = current_sring_buffers[device->index][device->tx_index];
    current_sring_buffers[device->index][device->tx_index] = frame_buffer;

	// Was there an error the last time to use this tx desc?
	check_tx_desc(device);

/*
 *  Now set up the transmit pointer in the descriptor.
 */
    sring[device->index][device->tx_index].tdes2 = (U32) SAr_Mem2Pci((void*)frame_buffer, (PU32)&status);
	if (OK != status) {
		return ERR_MEM2PCIFAIL;
	}

#if CACHE_SUPPORTED
  			Cache_IBR((char*)frame_buffer, ((char*)frame_buffer)+ETHER_BUFFER_SIZE);
#endif

/* 
 * Fill in fields in buffer descriptor and give buffer to the 21040.
 * (By changing the ownership bit).  The mb before the ownership
 * bit is changed is to make sure the other memory writes happen
 * and the descriptor is correctly set up before the ownership is
 * changed.
 */
    sring[device->index][device->tx_index].u_tdes1.tdes1 = 0;
    sring[device->index][device->tx_index].u_tdes1.s_tdes1.TBS1 = size;
    sring[device->index][device->tx_index].u_tdes1.s_tdes1.FS = TRUE;
    sring[device->index][device->tx_index].u_tdes1.s_tdes1.LS = TRUE;
    sring[device->index][device->tx_index].u_tdes1.s_tdes1.AC = FALSE;
    if (device->tx_index == (SRING_SIZE - 1))
		sring[device->index][device->tx_index].u_tdes1.s_tdes1.TER = TRUE;

    sring[device->index][device->tx_index].u_tdes0.tdes0 = 0;

    sring[device->index][device->tx_index].u_tdes0.s_tdes0.OWN = TRUE;
/*
 *  Advance the transmit ring index.
 */
    device->tx_index = (device->tx_index + 1) % SRING_SIZE;
/* 
 *  If the transmit process has stopped or is suspended, then boot it with a size 10.
 */
    csr5 = _21x4xReadCSR(device, CSR5);
    if (((csr5 & CSR5_TPS) != 0) || ((csr5 & CSR5_TS) == CSR5_TS_SUSPENDED)) {
		_21x4xWriteCSR(device, CSR1, CSR1_TPD);
    }
	
    return 0;
}
/* DECchip_21x4x_device_flush - wait for all writes on an ethernet device to complete
 *
 * Argument:
 *
 *    device_no - device to be flushed
 *
 * Returned value:
 *    TRUE	    - flush successful
 *    FALSE	    - flush timeout failure
 */

static int DECchip_21x4x_device_flush(int device_no)
{
device_no = device_no;
    return 0;

}
/* DECchip_21x4x_device_print_stats - print device statistics
 *
 * Argument:
 *
 *     device_no - device number
 */
static void DECchip_21x4x_device_print_stats(int device_no)
{
device_no = device_no;
    return;
}

/* DECchip_21x4x_device_preg - print all 21040 registers
 *
 * Argument:
 *
 *     device_no - device number
 */

static void DECchip_21x4x_device_preg(int device_no)
{
device_no = device_no;
    return;

}

/* DECchip_21x4x_device_clear_interrupts - clear all interrupts from an ethernet device
 *
 * Argument:
 *
 *     device_no - device number,
 */

static void DECchip_21x4x_device_clear_interrupts(int device_no)
{
    DECchip_21x4x_device_t *device = DECchip_21x4x_device_find(device_no);

/*
 *  Clear any interrupts there may be.
 */
    _21x4xWriteCSR(device, CSR5, _21x4xReadCSR(device, CSR5));

    return;
}

/* DECchip_21x4x_device_register - register a 21040 device driver in the 
 * Ethernet device table
 *
 * Arguments:
 *
 *    device_no - device number for registration
 *
 * Returned value
 *    TRUE         - registered successfully
 *    FALSE        - registration failed
 *
 * Description
 *
 *    Add a DECchip 21040 device to the installed driver list as follows:
 *    Check not all drivErs (DECCHIP21040_MAX_DEVICES) in use
 *    Initialise the appropriate current_buffers pointer array
 *    Set the appropriate driver_in_use flag
 *    Fill device number table with DECchip 21040 function pointers
 *		
 */
char *DECChip21040_description = "Digital Semiconductor 21040";
int DECchip_21x4x_device_register(int device_no)
{
  int (*get_hw_address_function)(int device_no, mac_addr hw_address);
  unsigned int pci;
  DECchip_21x4x_device_t *device = devices;
  int ID_Idx;
  int i;
  int devices_registered = 0;

/*
 *  The DBM will only support so many ethernet devices, so check.
 *  Only supporting 1 for StrongARM ebsa285 board
 */
  if (device_no >= MAX_ETHER_DEVICES) {
       return ERROR;
  }

  for (ID_Idx = 0; DC21X4X_DID_Data[ID_Idx].DeviceID != 0; ++ID_Idx) {
  	U32 inst=0;
	pci = SAr_FindPCIDevice((U16)DIGITAL_VID, (U16)DC21X4X_DID_Data[ID_Idx].DeviceID, &inst);
    if (pci != 0) {

/*
 *  By now all of our devices have been set up (by scanning the PCI
 *  slots and other devious means).  So, what we need to do is to
 *  assign the DBM device number to one of those (real) devices.
 *  Assuming, of course, that no one else has already allocated this
 *  device.  Check for this.
 */
		for (i = 0; i < DECCHIP21040_MAX_DEVICES; i++) {
			if (!(device->flags.allocated) && (device->flags.exists)) {
				device->flags.allocated = TRUE;
			/*
			 * Now that this device has been registered, give it a system wide
			 * device number.
			 */

			  device->exterior_device = device_no;
			if (DC21X4X_DID_Data[ID_Idx].DeviceID == DECCHIP_21040_DID)
			  	get_hw_address_function = DECchip_21040_device_get_hw_address;
			else
			    get_hw_address_function = DECchip_21x4x_device_get_hw_address;
				
			/*
			 * Register the device with the device type independent device
			 * register 
			 */
			ether_device_register(device_no, DECchip_21x4x_device_init, 
						DECchip_21x4x_device_read, DECchip_21x4x_device_write, 
						DECchip_21x4x_device_flush, DECchip_21x4x_device_print_stats, 
						DECchip_21x4x_device_preg, 
						DECchip_21x4x_device_clear_interrupts, 
						get_hw_address_function,
						//DECchip_21x4x_device_stop, // No stop in this version
						DECChip21040_description);
			  ++device_no;
			  ++devices_registered;
			  ++device;
			  break;
			}
	   }
	}
  }
  return(devices_registered);
}
/*
 * DECchip_21x4x_device_check_romid - check that a rom id is valid.
 *
 * Arguments:
 *    ROM ID - 32 bytes as follows:
 *  --------------------------------------------------
 *  bytes   contents
 *  --------------------------------------------------
 *  0-5     Address octets  0-5
 *  6-7     Checksum octets 1-2
 *  8-9     Checksum octets 2-1
 *  10-15   Address octets  5-0
 *  16-21   Address octets  0-5
 *  22-23   Checksum octets 1-2
 *  24-31   test pattern (ff-00-55-aa-ff-00-55-aa)
 *  --------------------------------------------------
 *
 * return:
 *    true if successful, false otherwise.
 */
#define ROM_SIZE 32
#define ROM_TEST_OFFSET	 24
#define ROM_TEST_SIZE 8
#define ROM_CHECKSUM_OFFSET 6
#define ROM_MAC_SIZE 6
static int DECchip_21x4x_device_check_romid(unsigned char *romid)
{
  unsigned char test_pattern[ROM_TEST_SIZE] = {0xff, 0x00, 0x55, 0xaa, 
						 0xff, 0x00, 0x55, 0xaa};
  U32 i;
/*
 *  first look at the test pattern.
 */

    for (i = 0; i < ROM_TEST_SIZE; i++)
	if (test_pattern[i] != romid[ROM_TEST_OFFSET + i]) {
	    return FALSE;
	}

/*
 *  We passed all of the tests, so return TRUE.
 */
    return TRUE;
}


/* DECchip_21040_device_get_hw_address - get the hardware address of a device 
 *
 * Arguments:
 *    device_no - device number for which hardware address is required
 *    hw_address - returned hardware mac address 
 */
static int DECchip_21040_device_get_hw_address(int device_no, 
  mac_addr hw_address)
{
    DECchip_21x4x_device_t *device = DECchip_21x4x_device_find(device_no);
    int i;
    U32 csr9;
    unsigned char romid[ROM_SIZE];

    if (device == NULL) {
 		return ERROR;
    }


	/*
	 *  If the device id has been read already it has been
	 *  safely tucked away in the network device's data structure.
	 *  Don't incur the overhead again of reading it from the ROM.
	 */
    if (DECchip_21x4x_device_hw_address_valid(device->hw_address)) {
     	__rt_memcpy(hw_address, device->hw_address, ROM_MAC_SIZE);


      return 0;
    }

	/*
	 *  First try to read the SROM ID via CSR 9.  If this fails
	 *  then try the ROM ID.
	 */
    _21x4xWriteCSR(device, CSR9, 0xFFFFFFFF);
    csr9 = _21x4xReadCSR(device, CSR9);
 
	/*
	 *  The SROM id is valid, so read it.  Each time we read the
	 *  CSR we get another byte of address.  We've already read
	 *  it once.
	 */
	for (i = 0; i < ROM_SIZE; i++) {
	    while (csr9 & CSR9_DN) {
			csr9 = _21x4xReadCSR(device, CSR9);
	    }
	    romid[i] = csr9 & CSR9_DT;
	    csr9 = _21x4xReadCSR(device, CSR9);
	}

	/*
	 *  Is the SROM ID itself valid?
	 */
	if (DECchip_21x4x_device_check_romid(romid)) {
	    __rt_memcpy(hw_address, romid, ROM_MAC_SIZE);


	    return 0;
	}
    					/* end of srom id is valid */
    return ERROR;

}

/* DECchip_21x4x_device_stop - Stop the device (preempts continued DMA activitity)
 *
 * Argument:
 *
 *     device_no - device number
 */
#if 0
static void DECchip_21x4x_device_stop(int device_no)
{
    DECchip_21x4x_device_t *device = DECchip_21x4x_device_find(device_no);

/*
 *  First find the device data structure for this device.
 */

    if (device == NULL) {
		return;
    }
    
    _21x4xWriteCSR(device, CSR0, _21x4xReadCSR(device, CSR0) | 1 );
    return;
}
#endif


/*
 *  Allocate (and check) all the buffers required by this device.
 */

static void DECchip_21x4x_device_allocate_buffers(
  DECchip_21x4x_device_t *device)
{
    U32 index = device->index;
    U32 i;

/*
 *  Allocate receive and transmit descriptors and 
 *  buffers for this device.
 */

    rring[index] = (DECCHIP_21x4x_RX_BD *) net_buffer_alloc(
      sizeof(DECCHIP_21x4x_RX_BD) * RRING_SIZE);
	if (NULL == rring[index]) {
		while(1);
	}


    sring[index] = (DECCHIP_21x4x_TX_BD *) net_buffer_alloc(
      sizeof(DECCHIP_21x4x_TX_BD) * SRING_SIZE);
	if (NULL == sring[index]) {
		while(1);
		
	}

    for (i = 0; i < RRING_SIZE; i++) {
		current_rring_buffers[index][i] = (PU8)framemalloc(ETHER_BUFFER_SIZE);
    }

    for (i = 0; i < SRING_SIZE; i++) {
		current_sring_buffers[index][i] = (PU8)framemalloc(ETHER_BUFFER_SIZE);
  	}
    device->setup_frame = (PS8)framemalloc(ETHER_BUFFER_SIZE);
}
/* 
 * DECchip_21x4x_device_init_module - initialise the module 
 *	retrun 0 if a device is found, otherwise ERROR
 */
int DECchip_21x4x_device_init_module(void) 
{
  U32 index;
  DECchip_21x4x_device_t *device = devices;
  U32 pci;
  int ID_Idx;
  U32 inst=0;
  	
	/*
		Is there any PCI?
		check for bench bit, ie. bench jumper set on ebsa285
	*/
	if (SAr_PciBenchMode())
		return ERROR;

	/*
	 * Set all of the devices to !exist until we find one.
	 */

    for (index = 0; index < DECCHIP21040_MAX_DEVICES; index++)
	devices[index].flags.exists = FALSE;

	/*
	 *  Look for the first of the 21x4xs in the system.
	 */


   index = 0;
   for (ID_Idx = 0; DC21X4X_DID_Data[ID_Idx].DeviceID != 0; ID_Idx++) {
   pci = SAr_FindPCIDevice((U16)DIGITAL_VID, (U16)DC21X4X_DID_Data[ID_Idx].DeviceID, &inst);
inst = 0;
    if (pci != 0) {
		U32 bar = 0;
		
		//  This slot contains a DECchip 21x4x chip.  
		device->flags.exists = TRUE;
	    device->flags.allocated = FALSE;
	    device->flags.initialised = FALSE;
	    device->flags.setup_frame = SETUP_FRAME_NEEDED;
	    device->flags.hw_valid = FALSE;
	    device->index = index;
		// Unfortunately 21143 & 21142 have same CFID reg		
		if (DC21X4X_DID_Data[ID_Idx].DeviceID == DECCHIP_21142_DID) {
			U32 rev_no;
			S32 status;
			
			rev_no = (0xf0 & SAr_PciCfgReadLong(pci, CFRV, (PS32)&status)) >> 4;
			if (rev_no > 1) {
				ID_Idx++; // its a 21143
			} 			
		}

		device->device_type = ID_Idx;
	    device->pci = pci;
	    device->PCI_IO_Base = (unsigned int)
		  (SAr_GetNextIOBAR(pci, &bar) + PCI_IO);

		/*
	 	*  Now allocate receive and transmit descriptors and 
	 	*  buffers for this device.
	 	*/

	    DECchip_21x4x_device_allocate_buffers(device);

		// Angel only wants 1 device so leave
		return 0;
    }
  }
return ERROR;  
}

static int DECchip_21x4x_device_hw_address_valid(mac_addr hw_address)
{
  int i, j;
  j = 0;
  for (i = 0; i < MAC_ADDRESS_SIZE; i++) j = j | hw_address[i];
  if (j == 0)
	return FALSE;
	
  return TRUE;
}



static void send_to_srom( U32 addr, U32 command)
{

  *(U32 *)addr = (U32)(command|SROM_SR|SROM_WR);

  return;
}

static U32 get_from_srom(U32 addr)
{

  return (volatile)*(U32 *)addr;
}



#define BASE SROM_SR|SROM_WR

unsigned short ReadSrom( U32 devAdrs, U8 lineCnt)
{
    int		ix;		/* index register */
    int		intCnt;	/* address loop count */
    U16	tmpShort;
    U32	tmpLong; 

	if (lineCnt < 64) {
		intCnt = 6;
		lineCnt = lineCnt << 2;  /* Prepare lineCnt for processing */
    }
    else
		intCnt = 8;

    /* Command the Serial ROM to the correct Line */

    /* Preamble of Command */

    send_to_srom(devAdrs, 0x00000000 | BASE);	/* Command 1 */
    _usleep (1);
    send_to_srom(devAdrs, 0x00000001 | BASE);	/* Command 2 */
    _usleep (1);
    send_to_srom(devAdrs, 0x00000003 | BASE);	/* Command 3 */
    _usleep (1);
    send_to_srom(devAdrs, 0x00000001 | BASE);	/* Command 4 */
    _usleep (1);


    /* Command Phase */

    send_to_srom (devAdrs, 0x00000005 | BASE);	/* Command 5 */
    _usleep (1);
    send_to_srom (devAdrs, 0x00000007 | BASE);	/* Command 6 */
    _usleep (1);
    send_to_srom (devAdrs, 0x00000005 | BASE);	/* Command 7 */
    _usleep (1);
    send_to_srom (devAdrs, 0x00000007 | BASE);	/* Command 8 */
    _usleep (1);
    send_to_srom (devAdrs, 0x00000005 | BASE);	/* Command 9 */
    _usleep (1);
    send_to_srom (devAdrs, 0x00000001 | BASE);	/* Command 10 */
    _usleep (1);
    send_to_srom (devAdrs, 0x00000003 | BASE);	/* Command 11 */
    _usleep (1);
    send_to_srom (devAdrs, 0x00000001 | BASE);	/* Command 12 */
    _usleep (1);
    
    /* Address Phase */

    for (ix = 0; ix < intCnt; ix++){
	tmpLong = (lineCnt & 0x80) >> 5; /* Extract and move bit to bit 3) */

		/* Write 'read' command */

    	send_to_srom (devAdrs, tmpLong | 0x00000001 | BASE);/* Command 13 */
    	_usleep (1);
    	send_to_srom (devAdrs, tmpLong | 0x00000003 | BASE);/* Command 14 */
    	_usleep (1);
    	send_to_srom (devAdrs, tmpLong | 0x00000001 | BASE);/* Command 15 */
    	_usleep (1);

		lineCnt <<= 1;  
	}

    _usleep (1);

    /* Data Phase */

    tmpShort =0;
    for (ix = 15; ix >= 0; ix--) {
		/* Write the command */

    	send_to_srom (devAdrs, 0x00000003 | BASE);		/* Command 16 */
    	_usleep (1);

		/* Extract the data */

    	tmpShort |= (((get_from_srom (devAdrs) & 0x00000008) >> 3)
		       << ix); 				/* Command 17 */
    	_usleep (1);
    	send_to_srom (devAdrs, 0x00000001 | BASE);		/* Command 18 */
    	_usleep (1);

    }

    /* Finish up command */

    send_to_srom (devAdrs, 0x00000000);	/* Command 19 */
    _usleep (1);

    return (tmpShort);
}



/* DECchip_21x4x_device_get_hw_address - get the hardware address of a device 
 *
 * Arguments:
 *    device_no - device number for which hardware address is required
 *    hw_address - returned hardware mac address 
 */
 
static void read_entire_srom(DECchip_21x4x_device_t *device)
{
 U16 SromWord = 0;
 U32 addr;
 U16 *SromP;
 int i;
  
   	SromP = (U16 *)&Srom;
    addr = device->PCI_IO_Base + CSR9;
	
	for (i=0; i<(sizeof(struct de4x5_srom)/sizeof(short)); i++) { // Fill the data struct 
		SromWord = ReadSrom(addr, i);
		*SromP++ = SromWord;
	}
}

static int DECchip_21x4x_device_get_hw_address(int device_no, mac_addr hw_address)
{
  DECchip_21x4x_device_t *device = DECchip_21x4x_device_find(device_no);

  

  if (device == NULL) {
    return ERROR;
  }

/*
 *  If the device id has been read already it has been
 *  safely tucked away in the network device's data structure.
 *  Don't incur the overhead again of reading it from the ROM.
 */
 

  	if (DECchip_21x4x_device_hw_address_valid(device->hw_address)) {
    	__rt_memcpy(hw_address, device->hw_address, ROM_MAC_SIZE);
    	return 0;
  	}
	read_entire_srom(device);
  	__rt_memcpy(hw_address, Srom.hw_addr, ROM_MAC_SIZE);


 return 0;
}

#endif /* NEED21040 */






