/****************************************************************************
**                                                                          *
**  Copyright  1997						     								*
**  by Digital Equipment Corporation, Maynard, Massachusetts.		     	*
**                                                                          *
**  All Rights Reserved							     						*
**                                                                          *
**  Permission  is  hereby  granted  to  use, copy, modify and distribute   *
**  this  software  and  its  documentation,  in  both  source  code  and   *
**  object  code  form,  and without fee, for the purpose of distribution   *
**  of this software  or  modifications  of this software within products   *
**  incorporating  an  integrated   circuit  implementing  Digital's  		*
**  StrongARM architecture,  regardless  of the  source of such integrated 	*
**	circuit, provided that the  above copyright  notice and this permission *
**  notice appear  in  all copies,  and  that  the  name  of  Digital  		*
**  Equipment Corporation  not  be  used  in advertising or publicity 		*
**  pertaining to distribution of the  document  or  software without 		*
**  specific, written prior permission.							            *
**                                                                          *
**  Digital  Equipment  Corporation   disclaims  all   warranties  and/or   *
**  guarantees  with  regard  to  this  software,  including  all implied   *
**  warranties of fitness for  a  particular purpose and merchantability,   *
**  and makes  no  representations  regarding  the use of, or the results   *
**  of the use of, the software and documentation in terms of correctness,  *
**  accuracy,  reliability,  currentness  or  otherwise;  and you rely on   *
**  the software, documentation and results solely at your own risk.	    *
**                                                                          *
**  StrongARM is a trademark of Digital Equipment Corporation               *
**                                                                          *
*****************************************************************************/
#ifndef LINT
extern char *rcsid;
#endif


/*
 * MODULE DESCRIPTION:
 * 
 *     UDP protocol interface for EB64 monitor
 *
 * 
 */
#include "netparam.h"
#include "net_buff.h"
#include "address.h"
#include "udp.h"
#include "ip.h"
#include "ethernet.h"
#include "support.h"
#include "debugio.h"
#include "errors.h"
#include "devconf.h"

#if (ETHERNET_SUPPORTED == 1)


/*
 * UDP header structure (see RFC 768)
 */

#define UDPH_SOURCE_OFFSET 0
#define UDPH_DESTINATION_OFFSET (UDPH_SOURCE_OFFSET+UDP_PORT_SIZE)
#define UDPH_LENGTH_OFFSET      (UDPH_DESTINATION_OFFSET+UDP_PORT_SIZE)
#define UDPH_CHKSUM_OFFSET      (UDPH_LENGTH_OFFSET + 2)
#define UDPH_HEADER_SIZE        (UDPH_CHKSUM_OFFSET + 2)     


#define UDP_IP_PROTOCOL_ID 17

#define MAX_REGISTERED_UDP_PORTS 10

static struct {
    int in_use;
    udp_packet_handler processing_function;
    udp_port port;
    void *protocol_data;
} udp_table[MAX_ETHER_DEVICES][MAX_REGISTERED_UDP_PORTS];

/* UDP Statistics */
static struct {
    int frames_received;
    int bad_lengths;
    int bad_chksums;
    int forwarded;
    int send_requests;
    int bad_send_lengths;
} udp_stats[MAX_ETHER_DEVICES];

/*
 * Internal functions
 */

/* udp_input - UDP packet processing function
 *
 * Arguments to packet processing functions:
 *     protocol_data - protocol data passed to module when protocol registered
 *     device_no     - device on which packet was received;
 *     packet        - packet, with IP header removed;
 *     size          - size of packet;
 *     source        - source IP address of packet;
 *     destination   - IP address to which the packet was addressed; this is needed by UDP.
 *     frame_buffer  - original buffer containing packet.
 *     balance_buffer- buffer returned to maintain buffer balance.
 *
 */
static void udp_input(void *protocol_data, 
  int device_no, 
  unsigned char* packet, 
  int size, 
  ip_addr source, 
  ip_addr destination, 
  unsigned char* frame_buffer, 
  unsigned char* *balance_buffer)
{
    int i;
    int port_index;

    /* Extract the UDP size */
    int udp_size = ((*(unsigned char *)(packet + UDPH_LENGTH_OFFSET)) << 8) +
      *((unsigned char *)(packet + UDPH_LENGTH_OFFSET + 1));

    udp_stats[device_no].frames_received++;

    /* Initialise the returned buffer for error cases */

    *balance_buffer = frame_buffer;

    /* Check that the size is valid */
    if (udp_size < UDPH_HEADER_SIZE || udp_size > size) {
		return;
    }

    /* Check whether there is a checksum in the packet; a zero checksum means
     * not used */
    if (*((unsigned char *)(packet + UDPH_CHKSUM_OFFSET)) != 0 ||
      *((unsigned char *)(packet + UDPH_CHKSUM_OFFSET + 1)) != 0) {

		/* There is a checksum; check it.
	 To check the checksum the packet has
		 * a psuedo header added to it. The 16 bit ones' complement sum of the
		 * words (16 bit alligned 16 bit fields) of the complete packet with
		 * the psedo header on the front should be (+ or -) 0.
	 The psuedo
		 * header consists of the following fields:     IP source address (4
		 * bytes)     IP destination address (4 bytes)     Zero filler (1 byte)
		 *     IP protocol id (1 byte)     UDP packet length; including real
		 * header but not psedo header (2 bytes)
	 The ones' complemement sum is
		 * calculated by calculating the 32 bit unsigned sum of the words and
		 * then adding the top 16 bits to the bottom 16 bits.  Since we know
 		 * that the checksum is nonzero the result of this will never be zero
		 * so it is only compared with -0 (0xFFFF) to check the checksum. */

		unsigned int sum;
		int packet_index;

		/* Calculate the sum of the psuedo header */
		sum =
		  ((source[0] + source[2] + destination[0] + destination[2]) << 8) + 
		  source[1] + source[3] + destination[1] + destination[3] +
		  UDP_IP_PROTOCOL_ID + udp_size;

		/* Add in the real packet */
		for (packet_index = 0; packet_index < udp_size - 1;
		  		packet_index += 2) {
		    sum += ((*(unsigned char *)(packet + packet_index)) << 8) +
		    *((unsigned char *)(packet + packet_index + 1));
		}

		/* If the packet contains an odd number of bytes add in the last byte
		 */
		if (udp_size % 2 == 1) {
		    sum += *((unsigned char *)(packet + packet_index)) << 8;
		}
		
		/* Convert to a 1's complement sum and check it */
		sum = (sum & 0xffff) + (sum >> 16);
		sum += (sum >> 16);
		
		if (sum != 0xffff) {
		    /* The checksum is wrong */
		    udp_stats[device_no].bad_chksums++;
		 	return;
		}
    }

	

    /* Packet correct - find out if the port is in use */
    for (port_index = 0; port_index < MAX_REGISTERED_UDP_PORTS; port_index++) {
		if (udp_table[device_no][port_index].in_use && 
		  		__rt_memcmp((char*)&udp_table[device_no][port_index].port,
		 		(char*)(packet + UDPH_DESTINATION_OFFSET), UDP_PORT_SIZE) == 0) {

		    udp_port source_port;
			udp_port dest_port;
			
		    udp_stats[device_no].forwarded++;

		    /* Copy the source port so that it remains readable in the called
		     * routine even when the packet is reused */
		    for (i = 0; i < UDP_PORT_SIZE; i++) {
				source_port[i] =
				 *((unsigned char *)(packet + UDPH_SOURCE_OFFSET + i));
				dest_port[i] = 
				 *((unsigned char *)(packet + UDPH_DESTINATION_OFFSET + i));
			}
		
			
		    /* Call the packet handler bound to the port */
		    (*udp_table[device_no][port_index].processing_function)(
		      udp_table[device_no][port_index].protocol_data, 
		      device_no, 
		      packet + UDPH_HEADER_SIZE, 
		      udp_size - UDPH_HEADER_SIZE, 
		      source_port, 
			  dest_port,
		      source, 
		      frame_buffer, 
		      balance_buffer);

		    /* There can only be one entry per port so exit from the search */
		    break;
		}
    }

}

/* udp_register_port - internal function to register a port
 *
 * Arguments:
 *    device_no            - device for which the port is to be registered
 *    port                 - port to be registered
 *    processing_function  - function to process packets received on this port
 *    protocol_data        - arbitrary data to be passed back when a packet is received
 *
 * Returned value:
 *    0 - protocol registered
 *    -VE - failed to register protocol
 */

static int udp_register_port(int device_no, 
  udp_port port, 
  udp_packet_handler processing_function, 
  void *protocol_data)
{

    /* The function scans the table of port to see if this port is already
     * registered, if it is the entry is updated; if not it puts an entry for
     * the port in the first free slot. When a packet is received the table is 
     * scanned from begining to end so this should mean that the most commonly 
     * used ports (assumed to be the first ones registered) are found first. */
    int first_free_slot = MAX_REGISTERED_UDP_PORTS;
					/* Used to keep track of whether their 
					 * are any free slots and if so where; 
					 * the value MAX_REGISTERED_UDP_PORTS
					 * is used to mean that there are no
					 * free slots */
    int current_slot;

    /* Scan from end backwards to simplify setting first_free_slot */
    for (current_slot = MAX_REGISTERED_UDP_PORTS - 1; current_slot >= 0;
      current_slot--) {
		if (udp_table[device_no][current_slot].in_use) {
		    if (__rt_memcmp(udp_table[device_no][current_slot].port, port,
		     		 UDP_PORT_SIZE) == 0) {
				/* Entry already exists; fail */
				return ERR_UDPPORTEXISTS;
		    }
		} 
		else {
		   	first_free_slot = current_slot;
		}
    }

    /* If this point is reached a new entry is needed */
    if (first_free_slot != MAX_REGISTERED_UDP_PORTS) {
		/* There is a slot free; fill it in */
		__rt_memcpy(udp_table[device_no][first_free_slot].port, port,
		  UDP_PORT_SIZE);
		udp_table[device_no][first_free_slot].processing_function =
		  processing_function;
		udp_table[device_no][first_free_slot].protocol_data = protocol_data;
		udp_table[device_no][first_free_slot].in_use = TRUE;
		return 0;
    }
    else {
		return ERR_UDPNOFREESLOTS;
    }
}



/*
 * Externally callable functions
 */

/* udp_init - initialise udp protocol module
 *
 * Argument:
 *    device_no - device number on which udp should be enabled.
 */
void udp_init(int device_no)
{
    int port_index;

    /* Register the protocol with ip */
    ip_register_protocol(device_no, 
      UDP_IP_PROTOCOL_ID, 
      udp_input, 
      NULL);

    /* Clear out the device's port list */
    for (port_index = 0; port_index < MAX_REGISTERED_UDP_PORTS; port_index++)
	udp_table[device_no][port_index].in_use = FALSE;
    udp_stats[device_no].frames_received = 0;
    udp_stats[device_no].bad_lengths = 0;
    udp_stats[device_no].bad_chksums = 0;
    udp_stats[device_no].forwarded = 0;
    udp_stats[device_no].send_requests = 0;
    udp_stats[device_no].bad_send_lengths = 0;
}

/* udp_register_well_known_port - tells the udp module what to do with packets received for a particular
 *                                well known port number.
 *
 * Arguments:
 *
 *    device_no             - device number on which port should be registered.
 *    port                  - port to register.
 *    processing_function   - function to process received packets.
 *    protocol_data         - arbitrary data to be passed back when a packet is received
 *
 * Returned value:
 *    0   - protocol registered
 *    -VE  - unable to register protocol
 */
int udp_register_well_known_port(int device_no, 
  udp_port port, 
  udp_packet_handler processing_function, 
  void *protocol_data)
{
    return udp_register_port(device_no, port, processing_function,
      protocol_data);
}

/* udp_create_new_port - tells the udp module to create a new port; and what to do with data received
 *                       on that port.  The created port number will be outside the well known port 
 *                       number range.
 *
 * Arguments:
 *
 *    device_no             - device number on which port should be created.
 *    processing_function   - function to process received packets
 *    protocol_data         - arbitrary data to be passed back when a packet is received
 *    port                  - returned port number
 *
 * Returned value:
 *    0   - protocol registered
 *    -VE  - unable to register protocol
 */
int udp_create_port(int device_no, 
  udp_packet_handler processing_function, 
  void *protocol_data, 
  udp_port port)
{
    static unsigned int next_port = 0x8000;

    port[0] = next_port >> 8;
    port[1] = (next_port++) & 0xff;
    if (next_port == 0x10000)
	 next_port = 0x8000;
	
	 
    return udp_register_port(device_no, port, processing_function,
      protocol_data);
}

/* udp_delete_port - deletes a previously created port
 *
 * Argument:
 *    device_no   - device number on which port is registered.
 *    port        - port to be deleted.
 */
void udp_delete_port(int device_no, udp_port port)
{
    int port_index;

    /* Find the port */
    for (port_index = 0; port_index < MAX_REGISTERED_UDP_PORTS; port_index++) {
		if (udp_table[device_no][port_index].in_use && 
		  		__rt_memcmp(port, udp_table[device_no][port_index].port,
		  		UDP_PORT_SIZE) == 0) {
		    /* Port found; delete it */
		    udp_table[device_no][port_index].in_use = FALSE;
		    break;
		}
    }
}
/* udp_write - sends a packet on a UDP port
 *
 * Arguments:
 *
 *    device_no        - device on which to send the packet
 *    packet           - pointer to data packet to be sent
 *    destination_port - UDP port to which the packet should be sent
 *    destination_addr - IP address to which the packet should be sent
 *    source_port      - UDP source port
 *    size             - size of packet
 *    frame_buffer     - buffer containing packet
 *    balance_buffer   - buffer returned to maintain buffer balance
 *
 * Return values:
 *
 *    len  - send successfull
 *    -ve - Error occured
 *
 * DESCRIPTION:
 *    This function writes an udp data packet to the device.  To maintain the caller's 
 *    pool of buffers the driver must give the caller an unused buffer (balance_buffer) 
 *    in return for the buffer containing the frame to be transmitted.  
 *
 *    If the send succeeds then frame_buffer must not be accessed by the caller after this
 *    call.  If the send can't be queued then frame_buffer will remain valid and
 *    the returned value of balance_buffer is undefined and must not be used.
 *
 */
int udp_write(int device_no, 
  unsigned char* packet, 
  udp_port destination_port, 
  ip_addr destination, 
  udp_port source_port, 
  int size, 
  unsigned char* frame_buffer, 
  unsigned char* *balance_buffer)
{


    unsigned char* const udp_packet = packet - UDPH_HEADER_SIZE;
    int const udp_size = size + UDPH_HEADER_SIZE;
    unsigned int sum;
    int packet_index;
    ip_addr source;
    int i, ret;


	*balance_buffer = frame_buffer;
	
    // Check that there is enough space in the buffer 	
    if (udp_packet - frame_buffer < ip_data_offset()) {
		udp_stats[device_no].bad_send_lengths++;
		return ERR_UDPBUFFERTOOSMALL;
    }

    /* Fill in the header fields */
    for (i = 0; i < UDP_PORT_SIZE; i++)
		*(unsigned char *)(udp_packet + UDPH_SOURCE_OFFSET + i)= (unsigned char)source_port[i];
    for (i = 0; i < UDP_PORT_SIZE; i++)
		*(unsigned char *)(udp_packet + UDPH_DESTINATION_OFFSET + i)=(unsigned char )destination_port[i];
   	*(unsigned char *)(udp_packet + UDPH_LENGTH_OFFSET)= (unsigned char )(udp_size >> 8);
    *(unsigned char *)(udp_packet + UDPH_LENGTH_OFFSET + 1)=(unsigned char )udp_size & 0xFF;

    /* To calculate the checksum set the checksum field to zero */
    *(unsigned char *)(udp_packet + UDPH_CHKSUM_OFFSET)= 0;
    *(unsigned char *)(udp_packet + UDPH_CHKSUM_OFFSET + 1)= 0;

    /* Calculate the checksum To calculate the checksum the packet has a psedo
     * header added to it. The checksum is then the 16 bit ones' complement of 
     * the ones' complement sum of the words (16 bit alligned 16  bit fields)
     * of the complete packet with the psedo header on the front.  The psedo
     * header consists of the following fields:     IP source address (4 bytes)
     *     IP destination address (4 bytes)     Zero filler (1 byte)     IP
     * protocol id (1 byte)     UDP packet length; including real header but
     * not psedo header (2 bytes) The ones' complemement sum is calculated by 
     * calculating the 32 bit unsigned sum of the words and then adding the top
     * 16 bits to the bottom 16 bits.  If the result is +0 (0x0000) it is
     * converted to -0 (0xFFFF) since a checksum field of 0x0000 means no
     * checksum. */

    /* Get the source address */

    ip_get_device_address(device_no, source);

    /* Calculate the sum of the psedo header */

    sum = ((source[0] + source[2] + destination[0] + destination[2]) << 8) + 
      source[1] + source[3] + destination[1] + destination[3] +
      UDP_IP_PROTOCOL_ID + udp_size;
    /* Add in the real packet */
    for (packet_index = 0; packet_index < udp_size - 1; packet_index += 2) {
		sum += ( *(unsigned char *)(udp_packet + packet_index) << 8)
	  		+ *(unsigned char *)(udp_packet + packet_index + 1);
    }

    /* If the packet contains an odd number of bytes add in the last byte */
    if (udp_size % 2 == 1) {
		sum += *(unsigned char *)(udp_packet + packet_index) << 8;
    }

    /* Convert to a 1's complement sum */
    sum = (sum & 0xffff) + (sum >> 16);
    sum += (sum >> 16);

    /* Take its complement */
    sum = ~sum;

    /* Check for special case of +0 */
    if (sum == 0x0000)
	 	sum = 0xffff;

    /* Write it to the checksum field in the header */
    *(unsigned char *)(udp_packet + UDPH_CHKSUM_OFFSET)= (unsigned char)(sum >> 8);
    *(unsigned char *)(udp_packet + UDPH_CHKSUM_OFFSET + 1)= (unsigned char)(sum & 0xff);

    /* Packet complete; send it */

    ret = ip_write(device_no, udp_packet, destination, UDP_IP_PROTOCOL_ID,
      udp_size, frame_buffer, balance_buffer);
	return ret;	
}


/* udp_data_offset - returns the offset at which udp data should be placed in a new ethernet frame
 *
 * Return value:
 *    Offset.
 */
int udp_data_offset(void) 
{
    return ip_data_offset() + UDPH_HEADER_SIZE;
}
#endif
