/*****************************************************************************

       Copyright  1997 Digital Equipment Corporation,
                       Maynard, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, provided  
that the copyright notice and this permission notice appear in all copies  
of software and supporting documentation, and that the name of Digital not  
be used in advertising or publicity pertaining to distribution of the software 
without specific, written prior permission. Digital grants this permission 
provided that you prominently mark, as not part of the original, any 
modifications made to this software or documentation.

Digital Equipment Corporation disclaims all warranties and/or guarantees  
with regard to this software, including all implied warranties of fitness for 
a particular purpose and merchantability, and makes no representations 
regarding the use of, or the results of the use of, the software and 
documentation in terms of correctness, accuracy, reliability, currentness or
otherwise; and you rely on the software, documentation and results solely at 
your own risk. 

******************************************************************************

******************************************************************************/

#include "platform.h"
#include "post.h"

/*
 * Local storage (banner, tests/results database etc).
 */
char *Banner = "EBSA285 PowerOn Selftests (built " __DATE__ ", at " __TIME__ ")\n\r" ;

struct test {
  int (*testCode)(struct testResults *) ;
  char *name ;
} ;

struct test tests[] = {
  {0,          "Memory tests    "} ,
  {timerTests, "Timer tests     "} ,
  {flashTests, "Flash tests     "} ,
  {comTests,   "COM Port tests  "} ,
} ;

#define NUMBER_OF_TESTS (sizeof(tests)/sizeof(struct test))

void whoAmI(struct testResults *pResults)
{
  unsigned int t4 ;

  /* memorySize is already done */
  /* cpuId */
  pResults->cpuId = cpuIdRead() ;

  /* 21285 device and revision ids */
  pResults->deviceId = *(unsigned short *)(CSR_BASE + 0x02) ;
  pResults->revisionId = *(unsigned char *)(CSR_BASE + 0x08) ;

  /* flags is a lot of things, CFM/sw jumpers/X-Bus */
  pResults->flags = 0 ;

  /* image switch setting */
  t4 = *(unsigned int *)XBUS_DEV_2 ;
  t4 &= 0xF ;
  pResults->switchSetting = t4^0xF  ;

  /* Central Function Mode */
  t4 = GET32(CSR_BASE + SA_CONTROL) ;
  if (t4 & 0x80000000) pResults->flags |= FLAGS_CFM ;

  /* jumper settings */
  t4 = GET32(XBUS_DEV_2) ;
  t4 &= 0x70 ; 
  t4 = t4 >> 4 ;
  t4 ^= 0x7 ;
  pResults->flags |= t4 ; 
  
  /* cpuFrequency */
  pResults->cpuFrequency = calibrate(pResults) ;

  /* busFrequency */
  pResults->busFrequency = 0 ;
}

void putBanner(void) 
{
  PutString("\n\r==========================================================\n\r") ;
  PutString(Banner) ;
  PutString("==========================================================\n\r") ;
}

int printResultsSummary(struct testResults *pResults)
{
  int i ;
  int failed = 0 ;

  putBanner() ;

  /* EBSA-285 stuff */
  PutString("Memory size is ") ; PutHex(pResults->memorySize) ; PutString(", ") ;
  PutString("CPU id is ") ; PutHex(pResults->cpuId) ; PutString("\n\r") ;

  /* 21285 stuff */
  PutString("21285 device id is ") ; PutHex(pResults->deviceId) ; PutString(", ") ;
  PutString("21285 revision is ") ; PutHex(pResults->revisionId) ; PutString("\n\r") ;

  /* What frequency? */
  PutString("CPU Frequency is ") ;
  PutString(MHz(pResults->cpuFrequency)) ;
  PutString(" MHz\n\r") ;

  /* how are we running? */
  if ((pResults->flags & FLAGS_CFM) == 0) PutString("Non-") ;
  PutString("Central Function Mode, PCI ") ;
  if (pResults->flags & FLAGS_IBUF6) 
    PutString("enabled\n\r") ;
  else
    PutString("disabled\n\r") ;

  /* Print out the raw flags */
  PutString("Flags: ") ;
  PutString("J17 pins 9-10 ") ;
  if ((pResults->flags & FLAGS_IBUF6) == 0) PutString("not ") ;
  PutString("fitted, ") ;
  PutString("J17 pins 11-12 ") ;
  if ((pResults->flags & FLAGS_IBUF5) == 0) PutString("not ") ;
  PutString("fitted,\n\r") ;
  PutString("       J17 pins 13-14 ") ;
  if ((pResults->flags & FLAGS_IBUF6) == 0) PutString("not ") ;
  PutString("fitted ") ;

  PutString("\n\r") ;

  if (pResults->flashBase == 0) 
    PutString("Flash device not found/recognised\n\r") ; 
  else {
    PutString("Flash device at ") ; PutHex(pResults->flashBase) ;
    PutString(", ") ; PutHex(pResults->flashBlocks) ;
    PutString(" of size ") ; PutHex(pResults->flashBlockSize) ; 
    PutString("\n\r") ;
  }


  PutString("Image switch is ") ; PutHex(pResults->switchSetting) ; PutString("\n\r") ;

  PutString("\n\rTest Results Summary\n\r") ;
  for (i = 0; i < NUMBER_OF_TESTS; i++) {
    PutString("    ") ;
    PutString(tests[i].name) ;
    if (pResults->results[i].run == 1) {
      if (pResults->results[i].result == 0) 
	PutString("Successful") ;
      else {
	PutString("Failed") ;
	failed++ ;                /* at least one test failed */
      }
    } else {
      PutString("[Not Run]") ;
    }
    PutString("\n\r") ;
  }

  return failed ;
}

int main (int memorySize)
{
  int failed = 0 ;
  int i ;
  struct testResults *pResults;

  /* This image is running out of flash and so we must use someplace
     to save the results of the tests and any information that we
     were passed from the assembly code phase of these tests.  We
     use the top of memory less 1Mbyte. */
  pResults = (struct testResults *)((char *)memorySize - 0x100000) ; 

  /* mask all interrupts in the interrupt controller */
  PUT32(CSR_BASE + IRQ_ENABLE_CLEAR,0xFFFFFFFF);

  /* tell the world... */
  putBanner() ;

  /* Initialize:
     set up the test results data structure */
  for (i = 1; i < NUMBER_OF_TESTS; i++) {
    pResults->results[i].run = 0 ;
    pResults->results[i].result = -1 ;
  } 
  pResults->results[0].run = 1 ;            /* frig the memory tests */
  pResults->results[0].result = 0 ;

  /* work out who/what we are */
  pResults->memorySize = memorySize ;
  PutString("Using memory at ") ; PutHex((unsigned int)pResults) ; 
  PutString(" for results\n\r") ;
  whoAmI(pResults) ;

  /* do all of the tests, carry on even if we fail,
     we start from 1 'cos the memory tests have already
     been done. */
  PutString("-------\n\r") ;
  for (i = 1; i < NUMBER_OF_TESTS; i++) {
    pResults->results[i].result = (tests[i].testCode)(pResults) ;
    pResults->results[i].run = 1 ; 
    PutString("-------\n\r") ;
    if (pResults->results[i].result != 0) 
      SetLEDs(RED_LED) ;
  }

  /* print the results and return */
  failed = printResultsSummary(pResults) ;
  return failed;
}

